import scipy.io as sio
import vitalbci_funcs
import numpy as np

vp_fb_parameters = sio.loadmat('/home/bbci/data/results/studies/vitalbci_season1/vp_fb_parameters.mat')
performance = sio.loadmat('/home/bbci/data/results/studies/vitalbci_season1/performance.mat')
subdir_list = performance['subdir_list']
ivals = vp_fb_parameters['ival']
bands = vp_fb_parameters['band']
classes = vp_fb_parameters['classes']
Nsubj = len(subdir_list)
cvresultsBroader = [None] * Nsubj
epoCSPBroader = [None] * Nsubj
cvresultsBroader100 = [None] * Nsubj
epoCSPBroader100 = [None] * Nsubj
for i, fname in enumerate(subdir_list):
    fname = fname[0][0]
    subjname = fname.split('_')[0]
    fullfile = '/home/bbci/data/bbciMat/' + fname + '/imag_arrow' + subjname + '_250Hz.mat'  # _250Hz option
    ival = ivals[i].astype(int)
    band = bands[i]
    classi = [str(classes[i][0][0][0][0]), str(classes[i][0][0][1][0])]
    cvresultsBroader[i], epoCSPBroader[i] = vitalbci_funcs.load_and_clfy(fullfile, ival, [4, 40], classi, PSDfreq=[0, 100])
    cvresultsBroader100[i], epoCSPBroader100[i] = vitalbci_funcs.load_and_clfy(fullfile, ival, [4, 100], classi, PSDfreq=[0, 100])
    maxres = np.max([np.mean(cvresultsBroader[i]), np.mean(cvresultsBroader100[i])])
    print('Subject finished, maximum accuracy was %2.1f%% while original was %2.1f%%' % (100*maxres,
                                                                                     performance['acc_c'][0][i]))
BroaderWs = np.array([x.W for x in epoCSPBroader])
Broader100Ws = np.array([x.W for x in epoCSPBroader100])

BroaderAs = np.array([x.A for x in epoCSPBroader])
Broader100As = np.array([x.A for x in epoCSPBroader100])

Broaderscores = np.array([x.score for x in epoCSPBroader])
Broader100scores = np.array([x.score for x in epoCSPBroader100])

Broaderspecs = np.array([x.spec for x in epoCSPBroader])
Broader100specs = np.array([x.spec for x in epoCSPBroader100])

np.savez('/home/bbci/data/results/studies/vitalbci_season1/bbcpy_cv_results_broader', cvresultsBroader=cvresultsBroader,
         cvresultsBroader100=cvresultsBroader100, BroaderWs=BroaderWs, BroaderAs=BroaderAs, Broaderscores=Broaderscores,
         Broaderspecs=Broaderspecs, Broader100Ws=Broader100Ws, Broader100As=Broader100As,
         Broader100scores=Broader100scores, Broader100specs=Broader100specs)



import numpy as np
import scipy as sp
from numpy.lib.stride_tricks import sliding_window_view
from scipy import signal
from sklearn.base import BaseEstimator, TransformerMixin
import mne

from bbcpy.functions.base import ImportFunc


def lfilter(obj, band, axis=-1, order=5, filttype='*', filtfunc=sp.signal.butter):
    band = np.array(band)
    if len(band.shape):
        assert band.shape == (2,)
    if filttype == '*':
        if band.shape == (2,):
            filttype = 'bandpass'
        else:
            filttype = 'low'
    [b, a] = filtfunc(order, band / obj.fs * 2, filttype)
    return obj.__class__(sp.signal.lfilter(b, a, obj, axis=axis), *obj.__initargs__())

def lfilter_sos(obj, band, axis=-1, order=5, filttype='*', filtfunc=sp.signal.butter):
    band = np.array(band)
    if len(band.shape):
        assert band.shape == (2,)
    if filttype == '*':
        if band.shape == (2,):
            filttype = 'bandpass'
        else:
            filttype = 'low'
    sos = filtfunc(order, band / obj.fs * 2, filttype)
    return obj.__class__(signal.sosfilt(sos, obj, axis=axis), *obj.__initargs__())


#todo re-write as function? also could be done by lfilter
class NotchFilter(BaseEstimator, TransformerMixin):
    """Transformer for mne notch filtering.
    ----------
    sample_rate : int
        Signal sample rate.
    line_noise : int
        Line noise
    """

    def __init__(self, sample_rate, line_noise):
        self.sample_rate = sample_rate
        self.line_noise = line_noise

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        X_ = self._notch_filter(X)
        return X_

    def fit_transform(self, X, y=None):
        X_ = self._notch_filter(X)
        return X_

    def _notch_filter(self, dat_):
        filt = mne.filter.notch_filter(x=dat_.T,
                                       Fs=self.sample_rate,
                                       trans_bandwidth=7,
                                       freqs=np.arange(self.line_noise,
                                                       4*self.line_noise,
                                                       self.line_noise),
                                       fir_design='firwin',
                                       verbose=False,
                                       notch_widths=1,
                                       filter_length=dat_.shape[0] - 1)
        return filt.T


class BandPassFilter(BaseEstimator, TransformerMixin):
    """Wrapper for MNEs Band Pass filtering.

        Parameter
        ----------
        filter_bands : list
            List of filter bands.
        sample_rate : int
            Signal sample rate
        filter_len : int or str
            lenght of the filter. The default is 1000ms.
        l_trans_bandwidth : int, optional
            The default is 4.
        h_trans_bandwidth : int, optional
            The default is 4.
    """

    def __init__(self, *,
                 filter_bands,
                 sample_rate,
                 filter_len='1000ms',
                 l_trans_bandwidth=4,
                 h_trans_bandwidth=4):
        self.filter_bands = filter_bands
        self.sample_rate = sample_rate
        self.filter_len = filter_len
        self.l_trans_bandwidth = l_trans_bandwidth
        self.h_trans_bandwidth = h_trans_bandwidth
        self.filters = _calc_band_filters(self.filter_bands,
                                          self.sample_rate,
                                          self.filter_len,
                                          self.l_trans_bandwidth,
                                          self.h_trans_bandwidth)

    def fit(self, X, y=None):
        self.filters = _calc_band_filters(self.filter_bands,
                                          self.sample_rate,
                                          self.filter_len,
                                          self.l_trans_bandwidth,
                                          self.h_trans_bandwidth)
        return self

    def transform(self, X, y=None):
        X_ = _apply_filter(X, self.filters)
        return X_

    def fit_transform(self, X, y=None):
        self.filters = _calc_band_filters(self.filter_bands,
                                          self.sample_rate,
                                          self.filter_len,
                                          self.l_trans_bandwidth,
                                          self.h_trans_bandwidth)
        X_ = _apply_filter(X, self.filters)
        return X_


def _calc_band_filters(f_ranges,
                       sample_rate,
                       filter_len="1000ms",
                       l_trans_bandwidth=4,
                       h_trans_bandwidth=4):
    """Calculate band filters.

    This function returns for the given frequency band ranges filter
    coefficients with length "filter_len"
    Thus the filters can be sequentially used for band power estimation

    Parameters
    ----------
    f_ranges : TYPE
        DESCRIPTION.
    sample_rate : float
        sampling frequency.
    filter_len : int,
        lenght of the filter. The default is 1001.
    l_trans_bandwidth : TYPE, optional
        DESCRIPTION. The default is 4.
    h_trans_bandwidth : TYPE, optional
        DESCRIPTION. The default is 4.
    Returns
    -------
    filter_fun : array
        filter coefficients stored in rows.
    """
    filter_fun = []

    for a, f_range in enumerate(f_ranges):
        h = mne.filter.create_filter(None,
                                     sample_rate,
                                     l_freq=f_range[0],
                                     h_freq=f_range[1],
                                     fir_design='firwin',
                                     l_trans_bandwidth=l_trans_bandwidth,
                                     h_trans_bandwidth=h_trans_bandwidth,
                                     filter_length=filter_len,
                                     verbose=False)

        filter_fun.append(h)

    return np.array(filter_fun)


def _apply_filter(dat_, filter_fun):
    """
    For a given channel, apply previously calculated filters
    Parameters
    ----------
    dat_ : array (ns,)
        segment of data at a given channel and downsample index.
    sample_rate : float
        sampling frequency.
    filter_fun : array
        output of calc_band_filters.
    line_noise : int|float
        (in Hz) the line noise frequency.
    seglengths : list
        list of ints with the leght to which variance is calculated.
        Used only if variance is set to True.
    variance : bool,
        If True, return the variance of the filtered signal, else
        the filtered signal is returned.
    Returns
    -------
    filtered : array
        if variance is set to True: (nfb,) array with the resulted variance
        at each frequency band, where nfb is the number of filter bands used to decompose the signal
        if variance is set to False: (nfb, filter_len) array with the filtered signal
        at each freq band, where nfb is the number of filter bands used to decompose the signal
    """ #noqa
    filtered = []

    for filt in range(filter_fun.shape[0]):
        for ch in dat_.T:
            filtered.append(signal.convolve(ch, filter_fun[filt, :], mode='same'))

    return np.array(filtered).T

#todo re-write as function
class SlidingWindows(BaseEstimator, TransformerMixin):
    """Sliding window transformer class.

    Parameter
    ----------
    window_size : int
        Desired size of windows.
    step_size: int
        Step size in between consecutive windows. Must be 1 or larger.
    label_handling: {'unique', 'mean', 'max', 'min', 'first', 'last',
    'center'}
        Decides which label to assign to the extracted sliding windows.
    adjust_class_size: Bool, default: False
        Switch to adjust classes to have the same number of samples.
    max_samples_per_class: int, optional (default: -1)
        Maximum samples per class. Only used, if adjust_class_size=True.
        Can oversample.
        Uses size of largest class if set to -1.
        Uses size of smallest class if set to 0.
    feature_transformer: sklearn Transformer, optional (default: None)
        Sklearn transformer to call on sliding windows to return features.
    """
    def __init__(self,
                 *,
                 window_size=100,
                 window_size_test=None,
                 step_size=100,
                 step_size_test=None,
                 label_handling='unique',
                 label_handling_test=None,
                 adjust_class_size=False,
                 adjust_class_size_test=None,
                 max_samples_per_class=0,
                 max_samples_per_class_test=None,
                 feature_transformer=None):
        self.window_size = window_size
        self.window_size_test = window_size_test
        self.step_size = step_size
        self.step_size_test = step_size_test
        self.label_handling = label_handling
        self.label_handling_test = label_handling_test
        self.adjust_class_size = adjust_class_size
        self.adjust_class_size_test = adjust_class_size_test
        self.max_samples_per_class = max_samples_per_class
        self.max_samples_per_class_test = max_samples_per_class_test
        self.feature_transformer = feature_transformer
        self.class_count_test = None
        self.class_count_train = None

    def fit(self, X, y=None):
        """Fit (do nothing)."""
        return self

    def transform(self, X, y=None):
        """Transform X into sliding windows.

        Parameters
        ----------
        X : ndarray, shape (n_trials, n_channels)
            Multi-channel time-series.
        y : ndarray, shape (n_trials,)
            Label for each trial.
        Returns
        -------
        out :   ndarray, shape ((n_matrices-window_size)/step_size, n_channels,
                window_size)
            Sliding windows.
        """
        window_size = ifaisnonethenbelsea(self.window_size_test,
                                          self.window_size)
        step_size = ifaisnonethenbelsea(self.step_size_test, self.step_size)
        label_handling = ifaisnonethenbelsea(self.label_handling_test,
                                             self.label_handling)
        adj = ifaisnonethenbelsea(self.adjust_class_size_test,
                                  self.adjust_class_size)
        mspc = ifaisnonethenbelsea(self.max_samples_per_class_test,
                                   self.max_samples_per_class)

        res = sliding_windows(X,
                              y=y,
                              window_size=window_size,
                              step_size=step_size,
                              label_handling=label_handling,
                              adjust_class_size=adj,
                              max_samples_per_class=mspc)
        if y is None:
            return res

        Xt, yt = res
        self.class_count_test = np.unique(yt, return_counts=True)
        if self.feature_transformer is not None:
            Xt = self.feature_transformer.transform(Xt)
        return Xt, yt

    def fit_transform(self, X, y=None):
        res = sliding_windows(X,
                              y=y,
                              window_size=self.window_size,
                              step_size=self.step_size,
                              label_handling=self.label_handling,
                              adjust_class_size=self.adjust_class_size,
                              max_samples_per_class=self.max_samples_per_class)
        if y is None:
            return res

        Xt, yt = res
        self.class_count_train = np.unique(yt, return_counts=True)
        if self.feature_transformer is not None:
            Xt = self.feature_transformer.fit_transform(Xt, yt)
        return Xt, yt


def ifaisnonethenbelsea(a, b):
    """Return a if a is not None, else return b

    Parameters
    ----------
    a : Object
    b : Object

    Returns
    -------
    out : Object
    """
    if a is None:
        return b
    else:
        return a

def sliding_windows(X,
                    y=None,
                    *,
                    window_size=100,
                    step_size=100,
                    label_handling='unique',
                    adjust_class_size=False,
                    max_samples_per_class=-1):

    x_windowed = sliding_window_view(X,
                                     window_size,
                                     axis=0)[::step_size]
    if y is None:
        return x_windowed

    y_windowed = sliding_window_view(y,
                                     window_size,
                                     axis=0)[::step_size]

    Xt, yt = label_window_decision(y_windowed,
                                   x_windowed,
                                   label_handling)
    if adjust_class_size:
        idx = adjust_class_sizes(yt, max_samples_per_class)
        Xt, yt = Xt[idx], yt[idx]

    return Xt, yt


def adjust_class_sizes(y, max_samples_per_class=0):
    """Equalize class sizes.

        Parameters
        ----------
        y : ndarray, shape (n_trials,)
            Label vector.
        max_samples_per_class : int, default: 0
            Maximum number of samples that are extracted per class. Allows for
            oversampling. If -1, the maximum number of samples for any class
            is used. If 0, the minimum number of samples for any class is used.

        Returns
        -------
        idx :   ndarray, shape (max_samples_per_class*n_classes)
            Indices of elements in y to be extracted to adjust the class sizes.
    """
    classes, counts = np.unique(y, return_counts=True)
    if max_samples_per_class == -1:
        max_samples_per_class = np.max(counts)
    if max_samples_per_class == 0:
        max_samples_per_class = np.min(counts)
    full_idx = np.arange(0, len(y))
    reduced_idx = []
    for cl in classes:
        cl_idx = full_idx[y == cl]
        idx_subset = np.round(np.linspace(-0.5,
                                          len(cl_idx) - .501,
                                          max_samples_per_class)
                              ).astype(int)
        cl_idx_subset = cl_idx[idx_subset]
        reduced_idx.append(cl_idx_subset)
    reduced_idx = np.sort(np.concatenate(reduced_idx))
    return reduced_idx


def label_window_decision(y_windows, x_windows, handler='unique'):
    """Decide how to handle labels in windows.
        Parameter
        ----------
        y_windows : ndarray, shape (??)
            Label windows.
        x_windows: ndarray, shape (??)
            Data Windows.
        handler: {'unique', 'mean', 'max', 'min', 'first', 'last', 'center'}
            Decides which label to assign to the extracted sliding windows.
    """
    if handler == 'unique':
        # all labels are set to at least 1, otherwise it would fail for negative
        # labels
        y2 = y_windows + y_windows.min() + 1
        # substract first entry from full windows for each window then take sum
        unique_label = np.sum((y2.T - y2[:, 0]), axis=0)
        # if the sum is 0, labels across window are the same
        return x_windows[unique_label == 0], y_windows[unique_label == 0, 0]

    elif handler == 'mean':
        return x_windows, np.mean(y_windows, axis=1)
    elif handler == 'max':
        return x_windows, np.max(y_windows, axis=1)
    elif handler == 'min':
        return x_windows, np.min(y_windows, axis=1)
    elif handler == 'first':
        return x_windows, y_windows[:, 0]
    elif handler == 'last':
        return x_windows, y_windows[:, -1]
    elif handler == 'center':
        midpoint = y_windows.shape[-1]//2
        return x_windows, y_windows[:, int(midpoint)]
    else:
        print('Handler method not known. Returning original array.')
        return x_windows, y_windows



import warnings

import numpy as np
import bbcpy
import h5py
import os


def dataset(fname):
    dataset_tmp = np.load(fname, allow_pickle=True)
    dataset = np.empty_like(bbcpy.datatypes.eeg.Data, shape=dataset_tmp.flatten().shape)
    for i, data in enumerate(np.nditer(dataset_tmp, ['refs_ok'])):
        data = dataset_tmp[i]
        dataset[i] = bbcpy.datatypes.eeg.Data(data['X'], data['fs'],
                                              bbcpy.datatypes.eeg.Marker(data['mrk_pos'], data['mrk_class'],
                                                                         data['mrk_className']),
                                              bbcpy.datatypes.eeg.Chans(data['clab'], data['mnt']))
    return np.reshape(dataset, dataset_tmp.shape)


def data(fname):
    '''
    Usage:
        X, fs, clab, mnt = load_data(fname)
        X, fs, clab, mnt, mrk_pos, mrk_class, mrk_className = load_data(fname)
    Parameters:
        fname: name of the data file
    Returns:
        X:    a 2D array of multi-channel timeseries (channels x samples), unit [uV]
        fs:   sampling frequency [Hz]
        clab: a 1D array of channel names  (channels)
        mnt:  a 2D array of channel coordinates (channels x 2)
              The electrode montage "mnt" holds the information of the
              2D projected positions of the channels, i.e. electrodes,
              on the scalp - seem from the top with nose up.
        mrk_pos:   a 1D array of marker positions (in samples)
        mrk_class: a 1D array that assigns markers to classes (0, 1)
        mrk_className: a list that assigns class names to classes
    '''
    _, file_extension = os.path.splitext(fname)
    if file_extension == '.mat':
        return bbci_mat(fname)
    data = np.load(fname, allow_pickle=True)
    if 'X' in data.keys():
        X = data['X']
    else:
        X = []
        warnings.warn('File without EEG data loaded.')
    fs = data['fs']
    clab = data['clab']
    clab = [x[0] for x in clab]
    mnt = data['mnt']
    if len(data.keys()) > 4:
        mrk_pos = data['mrk_pos']
        mrk_class = data['mrk_class']
        mrk_className = data['mrk_className']
        return bbcpy.datatypes.eeg.Data(X, fs, bbcpy.datatypes.eeg.Marker(mrk_pos, mrk_class, mrk_className, fs),
                                        bbcpy.datatypes.eeg.Chans(clab, mnt))
    else:
        return bbcpy.datatypes.eeg.Data(X, fs, chans=bbcpy.datatypes.eeg.Chans(clab, mnt))


def bbci_mat(fname, load_data=True):
    '''
    Usage:
        X, fs, clab, mnt = load_data(fname)
        X, fs, clab, mnt, mrk_pos, mrk_class, mrk_className = load_data(fname)
    Parameters:
        fname: name of the data file
    Returns:
        X:    a 2D array of multi-channel timeseries (channels x samples), unit [uV]
        fs:   sampling frequency [Hz]
        clab: a 1D array of channel names  (channels)
        mnt:  a 2D array of channel coordinates (channels x 2)
              The electrode montage "mnt" holds the information of the
              2D projected positions of the channels, i.e. electrodes,
              on the scalp - seem from the top with nose up.
        mrk_pos:   a 1D array of marker positions (in samples)
        mrk_class: a 1D array that assigns markers to classes (0, 1)
        mrk_className: a list that assigns class names to classes
    '''
    from os.path import exists
    if not exists(fname):
        raise FileNotFoundError(
            'File: ' + fname + ' not found. Please check correct name, path and existence.') from OSError
    try:
        import scipy.io as sio
        data = sio.loadmat(fname, uint16_codec='ascii')
        chankeys = [k for k in data.keys() if k[:2] == 'ch']
        channums = np.array([int(k[2:]) for k in chankeys])
        X = np.array([data[k] for k in chankeys])[channums - 1, :, 0]
        fs = data['nfo']['fs'][0, 0][0][0]
        clab = [x[0] for x in data['nfo']['clab'][0, 0][0, :]]
        mnt = np.array([data['mnt']['x'][0, 0], data['mnt']['y'][0, 0]])[:, :, 0].T
        if 'mrk' in data.keys():
            if 'time' in data['mrk'].dtype.names:
                mrk_pos = data['mrk']['time'][0, 0][0, :]
                mrk_fs = 1000  # time in ms
            else:
                mrk_pos = data['mrk']['pos'][0, 0][0, :]
                mrk_fs = fs
            mrk_class = np.where(data['mrk']['y'][0, 0].T)[1]
            mrk_className = np.array(data['mrk']['className'][0, 0][0, :])
    except NotImplementedError:  # v7.3
        try:
            import h5py
            data = h5py.File(fname, 'r')
            chankeys = [k for k in data.keys() if k[:2] == 'ch']
            channums = np.array([int(k[2:]) for k in chankeys])
            X = np.array([data[k] for k in chankeys])[channums - 1, 0, :]
            fs = data['nfo']['fs'][0, 0]
            clab = [''.join([chr(int(x)) for x in y]) for y in [data[ref] for ref in data['nfo']['clab'][:, 0]]]
            mnt = np.array([data['mnt']['x'][0, :], data['mnt']['y'][0, :]]).T
            if 'mrk' in data.keys():
                if 'time' in data['mrk'].keys():
                    mrk_pos = data['mrk']['time'][:, 0]
                    mrk_fs = 1000
                else:
                    mrk_pos = data['mrk']['pos'][:, 0]
                    mrk_fs = fs
                # mrk_class = data['mrk']['y']
                mrk_class = np.where(data['mrk']['y'])[1]
                mrk_className = np.array(
                    [''.join([chr(int(x)) for x in y]) for y in [data[ref] for ref in data['nfo']['className'][:, 0]]])
            data.close()
        except:
            ValueError('Error reading ' + fname + '...')
            return
    if not load_data:
        X = []
    if 'mrk_pos' in locals():
        return bbcpy.datatypes.eeg.Data(X, fs, bbcpy.datatypes.eeg.Marker(mrk_pos, mrk_class, mrk_className, mrk_fs),
                                        bbcpy.datatypes.eeg.Chans(clab, mnt))
    else:
        return bbcpy.datatypes.eeg.Data(X, fs, chans=bbcpy.datatypes.eeg.Chans(clab, mnt))

def epo(fname):
    raise NotImplementedError('Epo loading is not yet implemented, sorry.')
    return bbcpy.datatypes.eeg.Epo(data(fname), data.__initargs___())



import bbcpy
from bbcpy.pipeline import make_pipeline
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as lda
import numpy as np
from sklearn.model_selection import cross_val_score, KFold, train_test_split
from sklearn import clone
from sklearn.metrics import roc_auc_score
from matplotlib import pyplot as plt
from matplotlib import patches


def load_and_clfy(fname, ival, band, classes, PSDfreq=[1, 40], cvSplits=5):
    print('Starting subj: ' + fname)
    print('band= %2.1f to %2.1fHz' % (band[0], band[1]))
    print('ival= %i to %ims' % (ival[0], ival[1]))
    print('classes= %s' % classes)

    data = bbcpy.load.eeg.bbci_mat(fname)
    data = data['~E*']  # exclude EMG and EOG
    # classinds = [np.where(data.mrk.className==classes[0]), np.where(data.mrk.className==classes[1])]
    data.mrk = data.mrk.select_classes(classes)
    cv = KFold(n_splits=cvSplits)

    epo = data.lfilter(band).epochs(ival)

    vary = bbcpy.functions.base.ImportFunc(np.var, axis=2)
    logy = bbcpy.functions.base.ImportFunc(np.log)
    spec = bbcpy.functions.spectral.PSD(freq_range=PSDfreq, nperseg=data.fs)

    CSPpipeline = make_pipeline(bbcpy.functions.spatial.CSP, vary, logy, lda())
    cvresult = cross_val_score(CSPpipeline, epo, data.mrk.y, cv=cv)

    bbcpy.functions.spatial.CSP.n_cmps = 8
    X_train, X_test, y_train, y_test = train_test_split(epo, epo.y, test_size=1 / cvSplits, random_state=0)
    epoCSP = clone(bbcpy.functions.spatial.CSP)
    epoCSP.fit(X_train, y_train)
    features = logy(vary(epoCSP.transform(X_train)))
    features_test = logy(vary(epoCSP.transform(X_test)))
    epoCSP.score = list([])
    # plt.figure()
    Nfeatures = features.shape[1]
    for i in range(Nfeatures):
        feature = features[:, i][:, np.newaxis]
        clf = lda().fit(feature, y_train)
        feature_test = features_test[:, i][:, np.newaxis]
        epoCSP.score.append(clf.score(feature_test, y_test))
        # plt.subplot(2, Nfeatures, i + 1)
        # bbcpy.visual.scalp.map(data, epoCSP.A[:, i], clim='sym', colorbar=False)
        # plt.title('Cmp  %i acc=%2.1f%%' % (i, 100 * epoCSP.score[i]))
        # sbpl2 = plt.subplot(2, Nfeatures, i + 1 + Nfeatures)
        # plt.plot(Pxx.t, Pxx[:, i, :].T)
        # bandrect = patches.Rectangle((band[0], Pxx.min()), np.diff(band), np.diff([Pxx.min(), Pxx.max()]),
        #                             facecolor='g')
        # sbpl2.add_patch(bandrect)

    epoCSP.score = np.array(epoCSP.score)
    epoCSP.spec = (spec(epoCSP.transform(data.epochs(ival))).classmean())

    print('band= %2.1f to %2.1fHz' % (band[0], band[1]))
    print(np.array2string(cvresult * 100, formatter={'float_kind': lambda x: "%2.1f%%" % x}))
    print('%2.1f%%' % (cvresult.mean() * 100))
    return cvresult, epoCSP


def load_and_clfy_Miklodyfire(fname, ival, band, classes, nCSP, cvSplits=5, scoring=None, excllev=1.4, estimator='scm'):
    print('Starting subj: ' + fname)
    print('band= %2.1f to %2.1fHz' % (band[0], band[1]))
    print('ival= %i to %ims' % (ival[0], ival[1]))
    print('classes= %s' % classes)

    data = bbcpy.load.eeg.data(fname)
    # data = data['~E*'] #exclude EMG and EOG
    data = data[['~E*', '~Fp*', '~AF*', '~FAF*', '~*9', '~*10', '~O*', '~I*', '~PO7', '~PO8']]
    # classinds = [np.where(data.mrk.className==classes[0]), np.where(data.mrk.className==classes[1])]
    data.mrk = data.mrk.select_classes(classes)
    cv = KFold(n_splits=cvSplits)

    epo = data.lfilter(band).epochs(ival)

    vary = bbcpy.functions.base.ImportFunc(np.var, axis=2)
    logy = bbcpy.functions.base.ImportFunc(np.log)

    CSPpipeline = make_pipeline(bbcpy.functions.spatial.CSP(n_cmps=nCSP), vary, logy, lda())
    cvresultCSP = cross_val_score(CSPpipeline, epo, data.mrk.y, cv=cv, scoring=scoring)

    CSPpipeline = make_pipeline(bbcpy.functions.spatial.CSP(n_cmps=nCSP, excllev=excllev, estimator=estimator), vary,
                                logy, lda())
    cvresultCSP2 = cross_val_score(CSPpipeline, epo, data.mrk.y, cv=cv, scoring=scoring)

    pipeline = make_pipeline(bbcpy.functions.classifiers.Wishart(excllev=excllev, estimator=estimator))
    cvresultwish = cross_val_score(pipeline, epo, data.mrk.y, cv=cv, scoring=scoring)

    pipeline = make_pipeline(bbcpy.functions.spatial.CSP(n_cmps=nCSP, excllev=excllev),
                             bbcpy.functions.classifiers.Wishart(excllev=excllev, estimator=estimator))
    cvresultCSPwish = cross_val_score(pipeline, epo, data.mrk.y, cv=cv, scoring=scoring)

    pipeline = make_pipeline(bbcpy.functions.classifiers.Miklodyfire(excllev=excllev, n_cmps=nCSP, estimator=estimator))
    cvresultMfire = cross_val_score(pipeline, epo, data.mrk.y, cv=cv, scoring=scoring)

    pipeline = make_pipeline(bbcpy.functions.classifiers.Miklodyfire(excllev=excllev, n_cmps=2, estimator=estimator))
    cvresultMfire_2 = cross_val_score(pipeline, epo, data.mrk.y, cv=cv, scoring=scoring)

    pipeline = make_pipeline(
        bbcpy.functions.classifiers.Miklodyfire2(excllev=excllev, n_cmps=nCSP, estimator=estimator))
    cvresultMfire2_1 = cross_val_score(pipeline, epo, data.mrk.y, cv=cv, scoring=scoring)

    pipeline = make_pipeline(bbcpy.functions.classifiers.Miklodyfire2(excllev=excllev, n_cmps=2, estimator=estimator))
    cvresultMfire2_2 = cross_val_score(pipeline, epo, data.mrk.y, cv=cv, scoring=scoring)

    print('band= %2.1f to %2.1fHz' % (band[0], band[1]))
    print(np.array2string(cvresultCSP * 100, formatter={'float_kind': lambda x: "%2.1f%%" % x}))
    print('%2.1f%%' % (cvresultCSP.mean() * 100))
    print(np.array2string(cvresultCSP2 * 100, formatter={'float_kind': lambda x: "%2.1f%%" % x}))
    print('%2.1f%%' % (cvresultCSP2.mean() * 100))
    print(np.array2string(cvresultwish * 100, formatter={'float_kind': lambda x: "%2.1f%%" % x}))
    print('%2.1f%%' % (cvresultwish.mean() * 100))
    print(np.array2string(cvresultCSPwish * 100, formatter={'float_kind': lambda x: "%2.1f%%" % x}))
    print('%2.1f%%' % (cvresultCSPwish.mean() * 100))
    print(np.array2string(cvresultMfire * 100, formatter={'float_kind': lambda x: "%2.1f%%" % x}))
    print('%2.1f%%' % (cvresultMfire.mean() * 100))
    print(np.array2string(cvresultMfire_2 * 100, formatter={'float_kind': lambda x: "%2.1f%%" % x}))
    print('%2.1f%%' % (cvresultMfire_2.mean() * 100))
    print(np.array2string(cvresultMfire2_1 * 100, formatter={'float_kind': lambda x: "%2.1f%%" % x}))
    print('%2.1f%%' % (cvresultMfire2_1.mean() * 100))
    print(np.array2string(cvresultMfire2_2 * 100, formatter={'float_kind': lambda x: "%2.1f%%" % x}))
    print('%2.1f%%' % (cvresultMfire2_2.mean() * 100))
    return [cvresultCSP, cvresultCSP2, cvresultwish, cvresultCSPwish, cvresultMfire, cvresultMfire_2, cvresultMfire2_1,
            cvresultMfire2_2]


def load_and_clfy_Miklodyfirefb(fname, fnamefb, ival, band, classes, nCSP, excllev=1.4, estimator='scm'):
    print('Starting subj: ' + fname)
    print('band= %2.1f to %2.1fHz' % (band[0], band[1]))
    print('ival= %i to %ims' % (ival[0], ival[1]))
    print('classes= %s' % classes)

    data = bbcpy.load.eeg.data(fname)
    data = data[['~E*', '~Fp*', '~AF*', '~FAF*', '~*9', '~*10', '~O*', '~I*', '~PO7', '~PO8']]
    data.mrk = data.mrk.select_classes(classes)
    epo = data.lfilter(band).epochs(ival)

    datafb = bbcpy.load.eeg.data(fnamefb)
    datafb = datafb[['~E*', '~Fp*', '~AF*', '~FAF*', '~*9', '~*10', '~O*', '~I*', '~PO7', '~PO8']]
    datafb.mrk = datafb.mrk.select_classes(classes)
    epofb = datafb.lfilter(band).epochs(ival)

    vary = bbcpy.functions.base.ImportFunc(np.var, axis=2)
    logy = bbcpy.functions.base.ImportFunc(np.log)

    CSPpipeline = make_pipeline(vary, logy, lda())
    CSPpipeline.fit(epo, epo.y)
    cvresultlda = [CSPpipeline.score(epofb, epofb.y),
                   roc_auc_score(epofb.y, CSPpipeline.predict(epofb))]

    CSPpipeline = make_pipeline(bbcpy.functions.spatial.CSP(n_cmps=2), vary, logy, lda())
    CSPpipeline.fit(epo, epo.y)
    cvresultCSP = [CSPpipeline.score(epofb, epofb.y),
                   roc_auc_score(epofb.y, CSPpipeline.predict(epofb))]

    CSPpipeline = make_pipeline(bbcpy.functions.spatial.CSP(n_cmps=2, excllev=excllev, estimator=estimator), vary,
                                logy, lda())
    CSPpipeline.fit(epo, epo.y)
    cvresultCSP2 = [CSPpipeline.score(epofb, epofb.y),
                    roc_auc_score(epofb.y, CSPpipeline.predict(epofb))]

    pipeline = make_pipeline(bbcpy.functions.spatial.CSP(n_cmps=2, excllev=excllev),
                             bbcpy.functions.classifiers.Wishart(excllev=excllev, estimator=estimator))
    pipeline.fit(epo, epo.y)
    cvresultCSPwish = [pipeline.score(epofb, epofb.y),
                       roc_auc_score(epofb.y, CSPpipeline.predict(epofb))]

    pipeline = make_pipeline(bbcpy.functions.classifiers.Miklodyfire(excllev=excllev, n_cmps=nCSP, estimator=estimator))
    pipeline.fit(epo, epo.y)
    cvresultMfire = [pipeline.score(epofb, epofb.y),
                     roc_auc_score(epofb.y, CSPpipeline.predict(epofb))]

    pipeline = make_pipeline(bbcpy.functions.classifiers.Miklodyfire(excllev=excllev, n_cmps=2, estimator=estimator))
    pipeline.fit(epo, epo.y)
    cvresultMfire_2 = [pipeline.score(epofb, epofb.y),
                       roc_auc_score(epofb.y, CSPpipeline.predict(epofb))]

    print('band= %2.1f to %2.1fHz' % (band[0], band[1]))
    print(np.array2string(cvresultlda[0] * 100, formatter={'float_kind': lambda x: "%2.1f%%" % x}))
    #print('%2.1f%%' % (cvresultlda[0].mean() * 100))
    print(np.array2string(cvresultCSP[0] * 100, formatter={'float_kind': lambda x: "%2.1f%%" % x}))
    #print('%2.1f%%' % (cvresultCSP[0].mean() * 100))
    print(np.array2string(cvresultCSP2[0] * 100, formatter={'float_kind': lambda x: "%2.1f%%" % x}))
    #print('%2.1f%%' % (cvresultCSP2[0].mean() * 100))
    print(np.array2string(cvresultCSPwish[0] * 100, formatter={'float_kind': lambda x: "%2.1f%%" % x}))
    #print('%2.1f%%' % (cvresultCSPwish[0].mean() * 100))
    print(np.array2string(cvresultMfire[0] * 100, formatter={'float_kind': lambda x: "%2.1f%%" % x}))
    #print('%2.1f%%' % (cvresultMfire[0].mean() * 100))
    print(np.array2string(cvresultMfire_2[0] * 100, formatter={'float_kind': lambda x: "%2.1f%%" % x}))
    #print('%2.1f%%' % (cvresultMfire_2[0].mean() * 100))
    return [cvresultlda, cvresultCSP, cvresultCSP2, cvresultCSPwish, cvresultMfire, cvresultMfire_2]


def plot_and_compare(scores, specs, performance, f=np.arange(2, 40)):
    bestcomps = scores.max(axis=1)
    bestcompsi = scores.argmax(axis=1)
    plt.figure()
    plt.subplot(221)
    plt.plot(performance['acc_c'][0], bestcomps * 100, '.')

    GAspec = np.array([spec[:, bestcompsi[i], :] for i, spec in enumerate(specs)]).mean(axis=0)

    plt.subplot(223)
    plt.plot(f, GAspec.T)

    normAvSpec = np.array(
        [spec[0, bestcompsi[i], :] - spec[1, bestcompsi[i], :] for i, spec in enumerate(specs)]).T
    normAvSpec /= -normAvSpec[np.abs(normAvSpec).argmax(axis=0), np.arange(normAvSpec.shape[1])]

    plt.subplot(122)
    plt.imshow(normAvSpec[:, bestcomps.argsort()[::-1]].T, cmap='jet', vmin=-1, vmax=1)
    sortedperform = (np.sort(bestcomps)[::-10] * 100).tolist()
    plt.yticks(np.arange(0, bestcomps.shape[0], 10), ['%2.1f%%' % x for x in sortedperform])
    plt.xticks(f[3::5] - 2, f[3::5])
    plt.xlabel('frequency [Hz]')
    plt.ylabel('accuracy')
    cbar = plt.colorbar()
    cbar.set_label('relative PSD difference between class 1 and 2')


def plot_and_compare_scalps(scores, As, dataset, classes, sorted=False):
    bestcomps = scores.max(axis=1) * 100
    if sorted:
        bestcompssorti = scores.max(axis=1).argsort()[::-1]
    else:
        bestcompssorti = np.arange(bestcomps.shape[0])
    bestcomps = bestcomps[bestcompssorti]
    bestcompsi = scores.argmax(axis=1)[bestcompssorti]

    plt.figure()
    nSubj = dataset.shape[0]
    nRow = int(np.round(np.sqrt(nSubj)))
    for i in range(nSubj):
        plt.subplot(nRow, nRow, i + 1)
        bbcpy.visual.scalp.map(dataset[bestcompssorti[i]], As[bestcompssorti[i]][:, bestcompsi[i]], clim='sym',
                               colorbar=False, senspos=False)
        classesi = classes[bestcompssorti[i]][0][0]
        plt.title('A=%2.1f%%, %s-%s' % (bestcomps[i], classesi[0][0][0].upper(), classesi[1][0][0].upper()))


def plot_and_compare_scalps_bands(Alphascores, Betascores, Thetascores, Broadscores, Broaderscores, Broader100scores,
                                  AlphaAs, BetaAs, ThetaAs, BroadAs, BroaderAs, Broader100As,
                                  Alphaspecs, Betaspecs, Thetaspecs, Broadspecs, Broaderspecs, Broader100specs,
                                  dataset, classes, performances, selection=None):
    import scipy.io as sio
    lf = sio.loadmat('C:/Users/Daniel Miklody/tubCloud/Artefact Headmodel/lf_vol_1922_NY_DiTripoleMuscle_SymEye.mat')
    Broader100tissues = search_sources(Broader100As, Broader100scores, dataset, lf['lf_trip'],
                                       lf['elec'][0][0][3],
                                       lf['sourcemodel'][0][0][0], lf['sourcemodel'][0][0][2])
    Broadertissues = search_sources(BroaderAs, Broaderscores, dataset, lf['lf_trip'], lf['elec'][0][0][3],
                                    lf['sourcemodel'][0][0][0], lf['sourcemodel'][0][0][2])
    Broadtissues = search_sources(BroadAs, Broadscores, dataset, lf['lf_trip'], lf['elec'][0][0][3],
                                  lf['sourcemodel'][0][0][0], lf['sourcemodel'][0][0][2])
    Alphatissues = search_sources(AlphaAs, Alphascores, dataset, lf['lf_trip'], lf['elec'][0][0][3],
                                  lf['sourcemodel'][0][0][0], lf['sourcemodel'][0][0][2])
    Betatissues = search_sources(BetaAs, Betascores, dataset, lf['lf_trip'], lf['elec'][0][0][3],
                                 lf['sourcemodel'][0][0][0], lf['sourcemodel'][0][0][2])
    Thetatissues = search_sources(ThetaAs, Thetascores, dataset, lf['lf_trip'], lf['elec'][0][0][3],
                                  lf['sourcemodel'][0][0][0], lf['sourcemodel'][0][0][2])
    plt.figure()
    if selection is None:
        nSubj = 10  # dataset.shape[0]
        selection = range(nSubj)
    else:
        nSubj = len(selection)
    nCol = 6
    for i, isubj in enumerate(selection):
        ax1 = plt.subplot(nSubj, nCol, nCol * i + 1)
        plot_and_compare_scalp_bands_specs_i(Thetascores[isubj], ThetaAs[isubj], Thetaspecs[isubj], Thetatissues[isubj],
                                             dataset[isubj], classes[isubj][0][0])
        plt.text(-0.25, 0.5, 'orig. %2.0f%%' % performances[i], horizontalalignment='right',
                 verticalalignment='center', transform=ax1.transAxes)

        plt.subplot(nSubj, nCol, nCol * i + 2)
        plot_and_compare_scalp_bands_specs_i(Alphascores[isubj], AlphaAs[isubj], Alphaspecs[isubj], Alphatissues[isubj],
                                             dataset[isubj], classes[isubj][0][0])

        plt.subplot(nSubj, nCol, nCol * i + 3)
        plot_and_compare_scalp_bands_specs_i(Betascores[isubj], BetaAs[isubj], Betaspecs[isubj], Betatissues[isubj],
                                             dataset[isubj], classes[isubj][0][0])

        plt.subplot(nSubj, nCol, nCol * i + 4)
        plot_and_compare_scalp_bands_specs_i(Broadscores[isubj], BroadAs[isubj], Broadspecs[isubj], Broadtissues[isubj],
                                             dataset[isubj], classes[isubj][0][0])

        plt.subplot(nSubj, nCol, nCol * i + 5)
        plot_and_compare_scalp_bands_specs_i(Broaderscores[isubj], BroaderAs[isubj], Broaderspecs[isubj],
                                             Broadertissues[isubj], dataset[isubj], classes[isubj][0][0],
                                             spec_plot_f=np.arange(1, 90))

        plt.subplot(nSubj, nCol, nCol * i + 6)
        plot_and_compare_scalp_bands_specs_i(Broader100scores[isubj], Broader100As[isubj], Broader100specs[isubj],
                                             Broader100tissues[isubj], dataset[isubj], classes[isubj][0][0],
                                             spec_plot_f=np.arange(1, 90))


def plot_and_compare_scalp_bands_specs_i(scores, As, spec, tissues, data, classes, f=None, spec_plot_f=None):
    if f is None:
        f = np.arange(spec.shape[2])
    if spec_plot_f is None:
        spec_plot_f = np.arange(f.shape[0])
    bestcomps = scores.max() * 100
    bestcompsi = scores.argmax()
    ymin = spec[:, bestcompsi, spec_plot_f].min()
    ymax = spec[:, bestcompsi, spec_plot_f].max()
    yextent = ymin  # (ymax-ymin)/4+ymin
    xextent = spec.shape[2] / 3 * 2  # -(ymax-ymin)/4*3
    xspan = spec.shape[2]
    bbcpy.visual.scalp.map(data, As[:, bestcompsi], clim='sym',
                           colorbar=False, senspos=False, aspect='auto', extent=[xextent, xspan, yextent, ymax])
    plt.plot(f[spec_plot_f], spec[:, bestcompsi, spec_plot_f].T)
    plt.axis('on')
    # plt.xlabel('frequency [Hz]')
    # plt.ylabel('PSD [dB]')
    plt.text(5, 0.7 * (ymax - ymin) + ymin, 'A=%2.1f%%, %s, %s-%s' % (bestcomps, tissues[:5], classes[0][0][0].upper(),
                                                                      classes[1][0][0].upper()))


def search_sources(As, scores, dataset, lf, lfelec, grid, gridtissues):
    bestcompsi = scores.argmax(axis=1)
    tissues = []
    for i in range(As.shape[0]):
        chaninds = np.array([dataset[i].chans.index(chan[0][0] , markmissing=True) for chan in lfelec])
        validchans = [not chanind is None for chanind in chaninds]
        s, vmax, imax, dip_mom, dip_loc = bbcpy.functions.sourcespace.music(
            As[i][chaninds[validchans].astype(int), bestcompsi[i]], lf[validchans], grid)
        tissues.append(gridtissues[imax][0][0])
    return tissues


def EPSP(fs, normalize=False):
    tau1 = 0.0001
    tau2 = 0.002
    t = np.arange(0, 0.1, 1 / fs)
    y = 7 * (-np.exp(-t / tau1) + np.exp(-t / tau2))
    if normalize:
        y /= np.abs(y).max()
    return t, y


def IPSP(fs, normalize=False):
    tau1 = 0.005
    tau2 = 0.01
    t = np.arange(0, 0.1, 1 / fs)
    y = -(-np.exp(-t / tau1) + np.exp(-t / tau2))
    if normalize:
        y /= np.abs(y).max()
    return t, y


def MUAP(fs, normalize=False):
    U = 4  # m/s
    t = np.arange(0, 0.1, 1 / fs)
    z = 2 * U * t * 1000 / 100 * 4
    y = 4 * 96 * z * (6 - 6 * z + z ** 2) * np.exp(-z)
    if normalize:
        y /= np.abs(y).max()
    return t, y


def fitPSD(pxx):
    neuronAP = np.load('../NEURONSomaAP.npz')
    fs1 = 1000  # dataset[0].fs
    fs2 = int(1 / (neuronAP['t'][1] - neuronAP['t'][0]))
    yV = neuronAP['V'] - neuronAP['V'][0]
    yV /= np.abs(yV).max()
    tI, yI = IPSP(fs1, True)
    tE, yE = EPSP(fs1, True)
    tM, yM = MUAP(fs2, True)
    f, pxx = welch(np.pad(yV, (0, int(fs2) - neuronAP['V'].shape[0])), fs2,
                   nperseg=int(fs2 / 2), window='hamming', detrend=False)
    fyE, pyyE = welch(np.pad(yE, (0, int(fs1) - yE.shape[0])), fs1,
                      nperseg=int(fs1 / 2), window='hamming', detrend=False)
    fyI, pyyI = welch(np.pad(yI, (0, int(fs1) - yI.shape[0])), fs1,
                      nperseg=int(fs1 / 2), window='hamming', detrend=False)
    fyM, pyyM = welch(np.pad(yM, (0, int(fs2) - yM.shape[0])), fs2,
                      nperseg=int(fs2 / 2), window='hamming', detrend=False)

    return y1, h1, h1ei, y2, h2, h2ei, m



__all__ = ["structural", "statistics", "temporal", "spatial", "normalizations", "spectral", "sourcespace", "base", "classifiers"]
from bbcpy.functions import structural

from bbcpy.functions import statistics

from bbcpy.functions import temporal

from bbcpy.functions import spatial

from bbcpy.functions import normalizations

from bbcpy.functions import spectral

from bbcpy.functions import sourcespace

from bbcpy.functions import base

from bbcpy.functions import classifiers



import numpy as np
from bbcpy.functions.base import ImportFunc


def baseline(epo, ref_ival):
    '''
    Usage:
        epo = baseline(epo, epo_t, ref_ival)
    Parameters:
        epo: a 3D array of segmented signals, see makeepochs
        epo_t: a 1D array of time points of epochs relative to marker (in ms)
        ref_ival: a two element vector specifying the time interval for which the baseline is calculated [ms]
    '''
    if ((ref_ival[0] < epo.t[0]) | (ref_ival[1] > epo.t[-1])):
        raise ValueError('reference interval not inside epo')
    idxref = (ref_ival[0] <= epo.t) & (epo.t <= ref_ival[1])
    eporef = np.mean(epo[:, :, idxref], axis=2, keepdims=True)
    epo = epo - eporef
    return epo


# Now building the function as a transformer type:
baseline = ImportFunc(baseline, ref_ival=[-100, 0])



import bbcpy
import numpy as np
from matplotlib import pyplot as plt
from sklearn.model_selection import KFold, ShuffleSplit, GridSearchCV
from bbcpy.pipeline import make_pipeline
from sklearn.model_selection import cross_val_score

cv = KFold()

ival = [1000, 4500]
band = [10.5, 13]
imagVPaw = bbcpy.load.eeg.data('C:/data/bbcpydata/imagVPaw.npz')

cnt_bp = imagVPaw.lfilter(band)
epo_bp = cnt_bp.epochs(ival)

wish = bbcpy.functions.classifiers.Wishart()
wish.fit(epo_bp, epo_bp.y)
y = wish.decision_function(epo_bp)
print(np.mean(y))
plt.figure()
plt.plot(y)
plt.stem(np.arange(epo_bp.y.shape[0]), y.mean() * (2 * epo_bp.y.astype(float) - 1))
from pyriemann.utils import covariance

wish_lwf = bbcpy.functions.classifiers.Wishart(estimator=covariance._lwf)
wish_lwf.fit(epo_bp, epo_bp.y)
y2 = wish_lwf.decision_function(epo_bp)
print(np.mean(y2))

wish_art = bbcpy.functions.classifiers.Wishart(excllev=1.2)
wish_art.fit(epo_bp, epo_bp.y)
y2 = wish_art.decision_function(epo_bp)
print(np.mean(y2))
plt.figure()
plt.plot(y)
plt.stem(np.arange(epo_bp.y.shape[0]), y.mean() * (2 * epo_bp.y.astype(float) - 1))

# parameters = {'excllev': np.concatenate([[None], np.arange(1, 2, 0.05)]), 'autobias': [True, False],
#              'estimator': [covariance._scm, covariance._lwf, covariance._oas]}
# clf = GridSearchCV(bbcpy.functions.classifiers.Wishart(), parameters, cv=cv)
# clf.fit(epo_bp, epo_bp.y)
# print(clf.cv_results_['mean_test_score'])
# print(clf.cv_results_['mean_test_score'].max())
# maxind = clf.cv_results_['mean_test_score'].argmax()
# print([clf.cv_results_['params'][i] for i in np.where(clf.cv_results_['rank_test_score'] == 1)[0]])

pipeline = make_pipeline(bbcpy.functions.spatial.CSP(n_cmps=4, excllev=1.5),
                         bbcpy.functions.classifiers.Wishart(excllev=1.5))
print(np.mean(cross_val_score(pipeline, epo_bp, epo_bp.y, cv=cv)))

# pipeline = make_pipeline(bbcpy.functions.classifiers.Wishart(geomean=True))
# print(np.mean(cross_val_score(pipeline, epo_bp, epo_bp.y, cv=cv)))

pipeline = make_pipeline(bbcpy.functions.classifiers.Miklodyfire(excllev=1.5, n_cmps=4))
print(np.mean(cross_val_score(pipeline, epo_bp, epo_bp.y, cv=cv)))

parameters = {'excllev': np.concatenate([[None], np.arange(1, 2, 0.1)]),
              'estimator': [covariance._scm, covariance._lwf], 'n_cmps': [2, 3, 4, 6, 8]}
clf = GridSearchCV(bbcpy.functions.classifiers.Miklodyfire(), parameters, cv=cv)
clf.fit(epo_bp, epo_bp.y)
print(clf.cv_results_['mean_test_score'])
print(clf.cv_results_['mean_test_score'].max())
# maxind = clf.cv_results_['mean_test_score'].argmax()
# print(clf.cv_results_['params'][maxind])
print([clf.cv_results_['params'][i] for i in np.where(clf.cv_results_['rank_test_score'] == 1)[0]])

mik = bbcpy.functions.classifiers.Miklodyfire2(excllev=1.5, n_cmps=4, autobias=False)
mik.fit(epo_bp, epo_bp.y)

pipeline = make_pipeline(bbcpy.functions.classifiers.Miklodyfire2(excllev=1.5, n_cmps=4, autobias=True))
print(np.mean(cross_val_score(pipeline, epo_bp, epo_bp.y, cv=cv)))

parameters = {'excllev': np.concatenate([[None], np.arange(1, 2, 0.05)]), 'autobias': [True, False],
              'estimator': [covariance._scm, covariance._lwf], 'n_cmps': [2, 3, 4, 6, 8]}
clf = GridSearchCV(bbcpy.functions.classifiers.Miklodyfire2(), parameters, cv=cv)
clf.fit(epo_bp, epo_bp.y)
# print(clf.cv_results_['mean_test_score'])
print(clf.cv_results_['mean_test_score'].max())
# maxind = clf.cv_results_['mean_test_score'].argmax()
# print(clf.cv_results_['params'][maxind])
print([clf.cv_results_['params'][i] for i in np.where(clf.cv_results_['rank_test_score'] == 1)[0]])

pipeline = make_pipeline(bbcpy.functions.spatial.CSP(), bbcpy.functions.classifiers.Wishart())

parameters = {'wishart__excllev': np.concatenate([[None], np.arange(1, 2, 0.1)]), 'wishart__autobias': [True, False],
              'wishart__estimator': [covariance._scm, covariance._lwf, covariance._oas], 'csp__n_cmps': [4, 6, 8, 10]}
clf = GridSearchCV(pipeline, parameters, cv=cv)
clf.fit(epo_bp, epo_bp.y)
print(clf.cv_results_['mean_test_score'])
print(clf.cv_results_['mean_test_score'].max())
# maxind = clf.cv_results_['mean_test_score'].argmax()
# print(clf.cv_results_['params'][maxind])
print([clf.cv_results_['params'][i] for i in np.where(clf.cv_results_['rank_test_score'] == 1)[0]])

from scipy.spatial import distance_matrix

Sigma_tr = epo_bp.cov(target='trial')
Sigma = epo_bp.cov(target='class')
x_n = np.random.randn(*epo_bp.shape[1:])
x_n += 1. / distance_matrix(epo_bp.chans.mnt, epo_bp.chans['C3'].mnt[np.newaxis]+[0, 0.1]) * np.sin(
    2 * np.pi * 10 * np.arange(epo_bp.shape[2]) / epo_bp.fs)
Sigma_tr += x_n @ x_n.T
wx = np.trace((np.linalg.pinv(Sigma[0]) - np.linalg.pinv(Sigma[1])) @ Sigma_tr, axis1=1, axis2=2)
#wx -= np.trace((np.linalg.pinv(Sigma[0]) - np.linalg.pinv(Sigma[1])) @ x_n @ x_n.T, axis1=0, axis2=1)
#wx -= np.trace((np.linalg.pinv(Sigma[0]) - np.linalg.pinv(Sigma[1])) @ np.mean(Sigma_tr, axis=0), axis1=0, axis2=1)
b = (np.log(np.linalg.det(Sigma[0])) - np.log(np.linalg.det(Sigma[1])))
n = epo_bp.shape[-1]
y = (n - 1) * wx + n * b

plt.figure()
plt.hist([y[epo_bp.y == 0], y[epo_bp.y == 1]], 50, density=True)

y_corrected = y - (n-1) * np.trace((np.linalg.pinv(Sigma[0]) - np.linalg.pinv(Sigma[1])) @
                                   (1/(Sigma_tr.shape[0]) * np.sum(Sigma_tr, axis=0)), axis1=0, axis2=1)
# problems: if x_n and epo_bp are correlated? uneven means in the classes?
plt.figure()
plt.hist([y_corrected[epo_bp.y == 0], y_corrected[epo_bp.y == 1]], 50, density=True)

Sigma_tr = epo_bp.cov(target='trial')
Sigma = epo_bp.cov(target='class')
Sigma_tr += Sigma[0] - Sigma[1]
wx = np.trace((np.linalg.pinv(Sigma[0]) - np.linalg.pinv(Sigma[1])) @ Sigma_tr, axis1=1, axis2=2)
b = (np.log(np.linalg.det(Sigma[0])) - np.log(np.linalg.det(Sigma[1])))
n = epo_bp.shape[-1]
y = (n - 1) * wx + n * b
plt.figure()
plt.hist([y[epo_bp.y == 0], y[epo_bp.y == 1]], 50, density=True)

csp = bbcpy.functions.spatial.CSP()
for X in [epo_bp, csp(epo_bp)[0]]:
    Sigma_trial = X.cov(target='trial')
    Sigma_ = X.cov(target='class')
    Sigmainv_ = [np.linalg.pinv(Sigma_[0]), np.linalg.pinv(Sigma_[1])]
    n = X.shape[2]
    px1 = (n - 1) * np.trace(Sigmainv_[0] @ Sigma_trial, axis1=1, axis2=2) + n * np.log(np.linalg.det(Sigma_[0]))
    px2 = (n - 1) * np.trace(Sigmainv_[1] @ Sigma_trial, axis1=1, axis2=2) + n * np.log(np.linalg.det(Sigma_[1]))

    plt.figure()
    plt.plot(px1, px2, color=[0.8] * 3, linestyle='dashed')
    plt.scatter(px1, px2, c=X.y)
    pmax = np.max([px1, px2])
    plt.plot([0, pmax], [0, pmax])

    plt.figure()
    plt.plot(px1)
    plt.plot(px2)

    # plt.figure()
    # plt.plot(np.exp(-px1), np.exp(-px2), color='grey', linestyle='dashed')
    # plt.scatter(np.exp(-px1), np.exp(-px2), c=X.y)

from sklearn import svm

pipeline = make_pipeline(bbcpy.functions.spatial.CSP(n_cmps=2), bbcpy.functions.classifiers.Wishart(excllev=1.4),
                         svm.SVC(C=1))
print(np.mean(cross_val_score(pipeline, epo_bp, epo_bp.y, cv=cv)))

pipeline = make_pipeline(bbcpy.functions.classifiers.Miklodyfire(excllev=1.4), svm.SVC(C=1))
print(np.mean(cross_val_score(pipeline, epo_bp, epo_bp.y, cv=cv)))

wish_art = bbcpy.functions.classifiers.Miklodyfire(excllev=1.5, n_cmps=4)
wish_art.fit(epo_bp, epo_bp.y)
y = wish_art.transform(epo_bp)
plt.figure()
plt.plot(y[:, 0], y[:, 1], color=[0.8] * 3, linestyle='dashed')
plt.scatter(y[:, 0], y[:, 1], c=epo_bp.y)



import numpy as np
import bbcpy
#from bbcpy.pipeline import ImportFunc

def makeepochs(x: bbcpy.datatypes.eeg.Data, ival: np.ndarray, mrk = None) -> bbcpy.datatypes.eeg.Epo:
    """
    Usage:
        makeepochs(X, ival)
    Parameters:
        x: 2D array of multi-channel timeseries (channels x samples) of type EEGdata
        ival: a two element vector giving the time interval relative to markers (in ms)
    Returns:
        epo: a 3D array of segmented signals (samples x channels x epochs) of type EEGepo
    """
    if isinstance(ival, slice):
        if ival.start is None:
            start = 0
        else:
            start = ival.start
        if ival.stop is None:
            stop = 0
        else:
            stop = ival.stop
        if not ival.step is None:
            x = x[::ival.step]
            start = start / ival.step
            stop = stop / ival.step
        ival = [start, stop]
    if mrk is None:
        mrk = x.mrk.copy()
    time = np.arange(int(np.floor(ival[0] * x.fs / 1000)),
                     int(np.ceil(ival[1] * x.fs / 1000)) + 1, dtype=int)[np.newaxis, :]
    T = time.shape[1]
    nEvents = len(mrk)
    nChans = x.shape[0]
    idx = (time.T + np.array([mrk.in_samples(x.fs)])).reshape(T * nEvents).astype(int)
    epo = np.array(x)[:, idx].reshape(nChans, T, nEvents)
    epo = np.transpose(epo, (2, 0, 1))
    epo_t = np.linspace(ival[0], ival[1], T)
    epo = bbcpy.datatypes.eeg.Epo(epo, epo_t, x.fs, mrk, x.chans)
    return epo



import sklearn as sk
import numpy as np
from scipy import signal
import bbcpy

#todo rerwrite as simple function not transformer
class PSD(sk.base.BaseEstimator, sk.base.TransformerMixin):
    """Return the PSD
    """

    def __call__(self, data: bbcpy.datatypes.eeg.Data):
        return self.transform(data)

    def __init__(self, freq_range=[], window='hann', nperseg=100, axis=-1, dB=True):
        self.window = window
        self.nperseg = nperseg
        self.axis = axis
        self.freq_range = freq_range
        self.dB = dB

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        f, Pxx = signal.welch(X, fs=X.fs, window=self.window, nperseg=self.nperseg, axis=self.axis)
        if len(self.freq_range):
            fsel = (f >= self.freq_range[0]) & (f <= self.freq_range[1])
            if self.axis == -1:
                Pxx = Pxx[:, :, fsel]
            elif (self.axis == 2) | (self.axis == -2):
                Pxx = Pxx[:, fsel]
            else:
                Pxx = Pxx[fsel]
            f = f[fsel]
        if self.dB:
            Pxx = 10 * np.log10(np.abs(Pxx))
        return X.__class__(Pxx, f, *X.__initargs__()[1:])



import bbcpy
import numpy as np
from matplotlib import pyplot as plt
from scipy import signal

ival = [1000, 4500]
band = [10.5, 13]
imagVPaw = bbcpy.load.eeg.data('C:/data/bbcpydata/imagVPaw.npz')

#examples for new advanced indexing:

#select all but some (here exclude all frontal and fronto-central) channels:
print(imagVPaw['~F*'].shape)
#can also be combined - selecting all frontal channels except for 7 and 8s e.g.:
print(imagVPaw[['F?', '~*7', '~*8']].shape)
print(imagVPaw[['F?,~*7,~*8']].chans)
#adding advanced slicing with units - e.g. 2s to 5s in steps of 100ms:
print(imagVPaw['~F*', '2s:5s:100ms'].shape)
#you may also select single time points:
print(imagVPaw[:, '2s,5s,100ms'].shape)
#the same can be separated in a list:
print(imagVPaw[:, ['2s', '5s', '100ms']].shape)

cnt_bp = imagVPaw.lfilter(band)
epo_bp = cnt_bp.epochs(ival)

#also works on epoched data, also with comma separation, each taken individually
print(epo_bp[:, ['C?,~*3,~*4', 'FC*,~*3,~*4']].chans)

hilbilly = bbcpy.functions.base.ImportFunc(signal.hilbert, outClass='same', axis=-1)
ERD1 = np.abs(hilbilly(cnt_bp)).epochs(ival)
plt.figure()
plt.subplot(131)
bbcpy.visual.scalp.map(imagVPaw, ERD1.classmean(0)[0, :, '3s'])
plt.subplot(132)
bbcpy.visual.scalp.map(imagVPaw, ERD1.classmean(1)[0, :, '3s'])
plt.subplot(133)
bbcpy.visual.scalp.map(imagVPaw, ERD1.classmean(1)[0, :, '3s'] - ERD1.classmean(0)[0, :, '3s'])
plt.figure()
plt.subplot(211)
plt.plot(ERD1.t, np.squeeze(ERD1.classmean()[:, 'C3']).T)
plt.subplot(212)
plt.plot(ERD1.t, np.squeeze(ERD1.classmean()[:, 'C4']).T)
plt.show()

from sklearn import svm
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as lda
from sklearn.model_selection import KFold, ShuffleSplit, GridSearchCV
from bbcpy.pipeline import make_pipeline
from sklearn.model_selection import cross_val_score

# cv = ShuffleSplit(n_splits=5, test_size=0.3, random_state=0)
cv = KFold()

vary = bbcpy.functions.base.ImportFunc(np.var, axis=2)
clf = make_pipeline(vary, svm.SVC(C=1))
print(cross_val_score(clf, imagVPaw.lfilter(band).epochs(ival), ERD1.y, cv=cv))

# bbcpy.functions.statistics.cov=bbcpy.pipeline.FuncClass(bbcpy.functions.statistics.cov,target='trial') #moved to statistics.py
flatfeature = bbcpy.functions.base.ImportFunc(lambda x: np.reshape(x, (x.shape[0], -1)))
clf = make_pipeline(bbcpy.functions.statistics.cov, flatfeature, svm.SVC(C=1))
print(cross_val_score(clf, epo_bp, epo_bp.y, cv=cv))

# print(bbcpy.functions.spatial.CSP()(imagVPaw.lfilter(band).epochs(ival)))

clf = make_pipeline(bbcpy.functions.spatial.CSP(), vary, np.log, lda())
benchmark_res = cross_val_score(clf, epo_bp, epo_bp.y, cv=cv)
print(benchmark_res)

clf = make_pipeline(bbcpy.functions.spatial.CSP(), vary, np.log, svm.SVC(C=1))
print(cross_val_score(clf, epo_bp, epo_bp.y, cv=cv))

meany = bbcpy.functions.base.ImportFunc(np.mean, axis=-1)
clf = make_pipeline(meany, svm.SVC(C=1))
print(cross_val_score(clf, ERD1, ERD1.y, cv=cv))

clf = make_pipeline(flatfeature, svm.SVC(C=1))
print(cross_val_score(clf, ERD1, ERD1.y, cv=cv))

# testing baselining (and that its not  a good idea, here)
ival2 = [-100, 5000]
selival = bbcpy.functions.base.ImportFunc(lambda x: x[:, :, (x.t >= ival[0]) & (x.t <= ival[1])])
ERD2 = np.abs(hilbilly(cnt_bp)).epochs(ival2)
clf = make_pipeline(bbcpy.functions.normalizations.baseline, meany, svm.SVC(C=1))
print(cross_val_score(clf, ERD2, ERD2.y, cv=cv))



import numpy as np
import pytest

from bbcpy.functions.temporal import adjust_class_sizes, sliding_windows


@pytest.mark.parametrize("n_repetitions", [10, 20, 30])
def test_adjust_class_sizes(n_repetitions):
    y = np.repeat([0, 1, 1, 2, 2, 2], n_repetitions)
    classes, count = np.unique(y, return_counts=True)
    n_classes = len(classes)
    n_extracted = n_repetitions*5

    res = adjust_class_sizes(y, n_extracted)
    assert res.shape == (n_classes*n_extracted,)

    res = adjust_class_sizes(y)
    assert res.shape == (n_repetitions*n_classes,)

    res = adjust_class_sizes(y, int(n_repetitions//2))
    assert res.shape == ((n_repetitions*n_classes)//2,)


@pytest.mark.parametrize("label_handler", ['unique', 'mean', 'max', 'min',
                                           'first', 'last', 'center'])
def test_sliding_windows_label_handler(label_handler):
    """Test Sliding Windows label handler parameter."""
    n_channels = 3
    step_size = 4
    window_size = 8
    n_samples = 100
    X = np.random.rand(n_samples, n_channels)
    y = np.repeat([0, 1, 2, 3, 3], n_samples//5)

    X_win, y_win = sliding_windows(X,
                                   y=y,
                                   step_size=step_size,
                                   window_size=window_size,
                                   label_handling=label_handler
                                   )
    if label_handler == 'unique':
        n_expected_samples = 21
    else:
        n_expected_samples = np.ceil(n_samples / step_size) - 1
    assert X_win.shape == (n_expected_samples,
                           n_channels,
                           window_size)


def test_sliding_windows_label_handler2():
    """Test Sliding Windows label handler parameter."""
    n_channels = 3
    step_size = 4
    window_size = 8
    n_samples = 100

    X = np.random.rand(n_samples, n_channels)
    y = np.repeat([0, 1, 2, 3, 3], n_samples//5)

    X_win, y_win = sliding_windows(X,
                                   y=y,
                                   step_size=step_size,
                                   window_size=window_size,
                                   adjust_class_size=True
                                   )
    classes, counts = np.unique(y_win, return_counts=True)
    assert np.array_equal(classes, np.unique(y))
    assert np.sum(np.diff(counts)) == 0



import numpy as np
import scipy as sp
import sklearn as sk
from sklearn.base import TransformerMixin, BaseEstimator

import bbcpy.datatypes.eeg
#from bbcpy.functions import ImportFunc
#from bbcpy.datatypes import eeg


class CSP(sk.base.BaseEstimator, sk.base.TransformerMixin):
    def __call__(self, epo: bbcpy.datatypes.eeg.Epo, n_cmps=6):
        transformer = self.__class__(n_cmps)
        transformer.fit(epo)
        return transformer.transform(epo), transformer

    def __init__(self, n_cmps=6, excllev=None, estimator='scm'):
        self.n_cmps = n_cmps
        self.excllev = excllev
        self.estimator = estimator

    def fit(self, x, y=None, n_cmps=None):
        '''Fit CSP'''
        if n_cmps is None:
            n_cmps = self.n_cmps
        if y is not None:
            x.y = y
            x.mrk.y = y
        if self.excllev is not None:
            Sigma_trial = x.cov(target='trial')  # no estimator here because on single trials, this will fuck up excllev
            covtr = np.trace(np.linalg.pinv(x.cov(target='all', estimator=self.estimator)) @ Sigma_trial, axis1=1,
                             axis2=2) / x.shape[1]
            sel_tr = covtr <= self.excllev
            covs = x[sel_tr].cov(target='class', estimator=self.estimator)
        else:
            covs = x.cov(target='class', estimator=self.estimator)
        d, W = sp.linalg.eigh(covs[0], covs[0] + covs[1])
        selected_csps = np.flipud(np.argsort(np.maximum(d, 1 - d)))[:n_cmps]
        self.d = d[selected_csps]
        self.W = W[:, selected_csps]
        covX = x.cov(target='all', estimator=self.estimator)
        self.A = covX@W@(W.T@covX@W)[:, selected_csps]
        return self

    def transform(self, x, y=None):
        """Apply CSP.

        Parameters
        ----------
        x : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series
        y : ndarray, shape (n_trials, 2)
            Marker positions and marker class.
        Returns
        -------
        out : ndarray, shape (n_matrices, n_csp, n_time)
            transformed
        """
        return self.W.T@x

class CSP_general(BaseEstimator, TransformerMixin):
    def __init__(self, n_cmps=6):
        self.n_cmps=n_cmps

    def fit(self, x, y=None):
        '''Fit CSP'''
        classes = np.unique(y)
        c1 = sk.covariance.OAS().fit(x[y == classes[0]]).covariance_
        c2 = sk.covariance.OAS().fit(x[y == classes[1]]).covariance_
        d, W = sp.linalg.eigh(c1, c1 + c2)
        selected_csps = np.flipud(np.argsort(np.maximum(d, 1 - d)))[:self.n_cmps]
        self.d = d[selected_csps]
        self.W = W[:, selected_csps]
        return self

    def transform(self, x, y=None):
        """Apply CSP.
        Parameters
        ----------
        x : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series
        y : ndarray, shape (n_trials, 2)
            Marker positions and marker class.
        Returns
        -------
        out : ndarray, shape (n_matrices, n_csp, n_time)
            transformed
        """
        return (self.W.T@x.T).T


class BandWiseCSP(BaseEstimator, TransformerMixin):
    """Calculates and applies MNE's CSP.

    Parameters
    ----------
    n_components : int
        Number of components per filtered data subset.
    block_size : int or list
        Block sizes for the individual filter bands. If int, blocks are assumed
        to be of the same size. If a list is passed, the block sized are the
        individual entries.
    """
    def __init__(self, n_components, block_size):
        self.n_components = n_components
        self.block_size = block_size

    def fit(self, X, y=None):
        if type(self.block_size) is int:
            self._blocks = [self.block_size for i in
                            range(int(X.shape[1] // self.block_size))]
        else:
            self._blocks = self.block_size
        self.csps = []
        start = 0
        for i in self._blocks:
            end = start + i
            self.csps.append(CSP_general(n_cmps=self.n_components))
            self.csps[-1].fit(X[:, start:end], y)
            start = end

        return self

    def transform(self, X):
        filtered_signal = np.zeros(
            (X.shape[0], self.n_components * len(self._blocks)))
        start = 0
        for i, (block, csp) in enumerate(zip(self._blocks, self.csps)):
            end = start + block
            filtered_signal[:, i * self.n_components:(i + 1) * self.n_components] = csp.transform( # noqa
                X[:, start:end])
            start = end

        return filtered_signal

# todo rerwrite as simple function not transformer
class Rereferencing(BaseEstimator, TransformerMixin):
    """Rereferencing.
        ----------
        method : str
            Rereferencing method.
        reref_idx : list
            list of rereferencing electrodes for each channel for custom
            rereferencing.
        """

    def __init__(self, method, reref_idx=None):
        self.method = method
        self.reref_idx = reref_idx
        self.reref_function = self.reref_methods[method]

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return self.reref_function(self, X)

    def fit_transform(self, X, y=None):
        return self.reref_function(self, X)

    def reref_avg(self, X, reref_idx=None):
        return (X.T - X.mean(axis=1)).T

    def reref_bipolar(self, X, reref_idx=None):
        X_ = np.roll(X, 1)
        return X - X_

    def reref_custom(self, X):
        X_T = X.T
        for i, x in enumerate(X_T):
            x -= np.mean(X[self.reref_idx[i]], axis=1)
        return X_T.T

    reref_methods = {
        'average': reref_avg,
        'bipolar': reref_bipolar,
        'custom': reref_custom
    }


def car(data):
    return data - np.mean(data, axis=1)


def laplace(data):
    if hasattr(data, 'chans') and hasattr(data.chans, 'mnt'):
        return
    else:
        return

def laplace_dan2D(data):
    if hasattr(data, 'chans') and hasattr(data.chans, 'mnt'):
        return
    else:
        return



from functools import partial

import numpy as np
from sklearn.model_selection._split import _BaseKFold
from sklearn.utils.validation import _num_samples
from sklearn.metrics import (
    r2_score,
    median_absolute_error,
    max_error,
    mean_absolute_error,
    mean_squared_error,
    mean_squared_log_error,
    mean_poisson_deviance,
    mean_gamma_deviance,
    accuracy_score,
    top_k_accuracy_score,
    roc_auc_score,
    average_precision_score,
    log_loss,
    balanced_accuracy_score,
    explained_variance_score,
    brier_score_loss,
    mean_absolute_percentage_error,
)


class TrialWiseKFold(_BaseKFold):
    """Trial wise KFold cross-validator

        Provides train/test indices to split data in train/test sets. Split
        dataset into k consecutive folds (without shuffling by default) to
        achieve an equal number of movement trials in each split.
        Each fold is then used once as a validation while the k - 1 remaining
        folds form the training set.
    """

    def __init__(self, n_splits=5, *, shuffle=False, random_state=None):
        super().__init__(n_splits=n_splits, shuffle=shuffle,
                         random_state=random_state)

    def _iter_test_indices(self, X, y, groups=None):
        n_samples = _num_samples(X)
        indices = np.arange(n_samples)

        labeldiff = np.argwhere(np.diff(y) != 0)[:, 0] + 1
        labeldiff = np.insert(labeldiff, 0, 0)
        labeldiff = np.append(labeldiff, n_samples)

        ld_low = labeldiff[::2]
        ld_high = labeldiff[1::2]

        idx_low = ld_low[
            np.floor(np.linspace(0, len(ld_low) - 1,
                                 self.n_splits + 1)).astype(int)]
        idx_up = ld_high[np.floor(
            np.linspace(0, len(ld_high) - 1, self.n_splits + 1)).astype(int)]

        cv_indices = (idx_low + idx_up) // 2
        cv_indices[0] = 0
        cv_indices[-1] = n_samples
        for i in range(self.n_splits):
            yield indices[cv_indices[i]:cv_indices[i + 1]]


def make_scorer(estimator, X, y, error_score, **kwargs):
    """Create label transforming pipeline scorer.

    Parameters
    ----------
    estimator : scikit-lean pipeline
    X : input data
    y : input labels
    error_score : callable
        error score method
    kwargs
        kwargs passed to error score method

    Returns
    -------
    scorer : callable
        scorer that transforms the labels before comparing y_true and y_pred
    """
    Xt, y_true = estimator[:-1].transform(X, y)
    y_pred = estimator[-1].predict(Xt)
    return error_score(y_true, y_pred, **kwargs)


# Standard regression scores
explained_variance_scorer = partial(make_scorer,
                                    error_score=explained_variance_score)
r2_scorer = partial(make_scorer, error_score=r2_score)
max_error_scorer = partial(make_scorer, error_score=max_error,
                           greater_is_better=False)
neg_mean_squared_error_scorer = partial(make_scorer,
                                        error_score=mean_squared_error,
                                        greater_is_better=False)
neg_mean_squared_log_error_scorer = partial(make_scorer,
                                            error_score=mean_squared_log_error,
                                            greater_is_better=False
)
neg_mean_absolute_error_scorer = partial(make_scorer,
                                         error_score=mean_absolute_error,
                                         greater_is_better=False
)
neg_mean_absolute_percentage_error_scorer = partial(
    make_scorer, error_score=mean_absolute_percentage_error,
    greater_is_better=False
)
neg_median_absolute_error_scorer = partial(make_scorer,
                                           error_score=median_absolute_error,
                                           greater_is_better=False
)
neg_root_mean_squared_error_scorer = partial(make_scorer,
                                             error_score=mean_squared_error,
                                             greater_is_better=False,
                                             squared=False
)
neg_mean_poisson_deviance_scorer = partial(make_scorer,
                                           error_score=mean_poisson_deviance,
                                           greater_is_better=False
)

neg_mean_gamma_deviance_scorer = partial(make_scorer,
                                         error_score=mean_gamma_deviance,
                                         greater_is_better=False
)

# Standard Classification Scores
accuracy_scorer = partial(make_scorer, error_score=accuracy_score)
balanced_accuracy_scorer = partial(make_scorer,
                                   error_score=balanced_accuracy_score)

# Score functions that need decision values
top_k_accuracy_scorer = partial(make_scorer,
                                error_score=top_k_accuracy_score,
                                greater_is_better=True,
                                needs_threshold=True
)
roc_auc_scorer = partial(make_scorer,
                         error_score=roc_auc_score,
                         greater_is_better=True,
                         needs_threshold=True
)
average_precision_scorer = partial(make_scorer,
                                   error_score=average_precision_score,
                                   needs_threshold=True)
roc_auc_ovo_scorer = partial(make_scorer, error_score=roc_auc_score,
                             needs_proba=True,
                                 multi_class="ovo")
roc_auc_ovo_weighted_scorer = partial(make_scorer,
                                      error_score=roc_auc_score,
                                      needs_proba=True,
                                      multi_class="ovo",
                                      average="weighted"
)
roc_auc_ovr_scorer = partial(make_scorer, error_score=roc_auc_score,
                             needs_proba=True,
                                 multi_class="ovr")
roc_auc_ovr_weighted_scorer = partial(make_scorer,
                                      error_score=roc_auc_score,
                                      needs_proba=True,
                                      multi_class="ovr",
                                      average="weighted"
)

# Score function for probabilistic classification
neg_log_loss_scorer = partial(make_scorer,
                              error_score=log_loss,
                              greater_is_better=False,
                              needs_proba=True)
neg_brier_score_scorer = partial(make_scorer,
                                 error_score=brier_score_loss,
                                 greater_is_better=False,
                                 needs_proba=True
)
brier_score_loss_scorer = partial(make_scorer,
                                  error_score=brier_score_loss,
                                  greater_is_better=False,
                                  needs_proba=True
)


SCORERS = dict(
    explained_variance=explained_variance_scorer,
    r2=r2_scorer,
    max_error=max_error_scorer,
    neg_median_absolute_error=neg_median_absolute_error_scorer,
    neg_mean_absolute_error=neg_mean_absolute_error_scorer,
    neg_mean_absolute_percentage_error=neg_mean_absolute_percentage_error_scorer,  # noqa
    neg_mean_squared_error=neg_mean_squared_error_scorer,
    neg_mean_squared_log_error=neg_mean_squared_log_error_scorer,
    neg_root_mean_squared_error=neg_root_mean_squared_error_scorer,
    neg_mean_poisson_deviance=neg_mean_poisson_deviance_scorer,
    neg_mean_gamma_deviance=neg_mean_gamma_deviance_scorer,
    accuracy=accuracy_scorer,
    top_k_accuracy=top_k_accuracy_scorer,
    roc_auc=roc_auc_scorer,
    roc_auc_ovr=roc_auc_ovr_scorer,
    roc_auc_ovo=roc_auc_ovo_scorer,
    roc_auc_ovr_weighted=roc_auc_ovr_weighted_scorer,
    roc_auc_ovo_weighted=roc_auc_ovo_weighted_scorer,
    balanced_accuracy=balanced_accuracy_scorer,
    average_precision=average_precision_scorer,
    neg_log_loss=neg_log_loss_scorer,
    neg_brier_score=neg_brier_score_scorer,
)



import scipy.io as sio
import vitalbci_funcs
import numpy as np

vp_fb_parameters = sio.loadmat('/home/bbci/data/results/studies/vitalbci_season1/vp_fb_parameters.mat')
performance = sio.loadmat('/home/bbci/data/results/studies/vitalbci_season1/performance.mat')
subdir_list = performance['subdir_list']
ivals = vp_fb_parameters['ival']
bands = vp_fb_parameters['band']
classes = vp_fb_parameters['classes']
# ival = [580, 4470]
# band = np.array([15, 23]) / 2
# band = [7.5, 40]
Nsubj = len(subdir_list)
cvresultsAlpha = [None] * Nsubj
epoCSPAlpha = [None] * Nsubj
cvresultsBeta = [None] * Nsubj
epoCSPBeta = [None] * Nsubj
cvresultsTheta = [None] * Nsubj
epoCSPTheta = [None] * Nsubj
cvresultsBroad = [None] * Nsubj
epoCSPBroad = [None] * Nsubj

for i, fname in enumerate(subdir_list):
    fname = fname[0][0]
    subjname = fname.split('_')[0]
    fullfile = '/home/bbci/data/bbciMat/' + fname + '/imag_arrow' + subjname + '.mat'  # _250Hz option
    ival = ivals[i].astype(int)
    band = bands[i]
    classi = [str(classes[i][0][0][0][0]), str(classes[i][0][0][1][0])]
    if band[0] < 13:
        cvresultsTheta[i], epoCSPTheta[i] = vitalbci_funcs.load_and_clfy(fullfile, ival, band / 2, classi)
        cvresultsAlpha[i], epoCSPAlpha[i] = vitalbci_funcs.load_and_clfy(fullfile, ival, band, classi)
        if band[1] > 22.5:
            bandmax = 45
        else:
            bandmax = 2 * band[1]
        cvresultsBeta[i], epoCSPBeta[i] = vitalbci_funcs.load_and_clfy(fullfile, ival, [band[0] * 2, bandmax], classi)
        cvresultsBroad[i], epoCSPBroad[i] = vitalbci_funcs.load_and_clfy(fullfile, ival, [band[0], bandmax], classi)
    if band[0] >= 13:
        cvresultsTheta[i], epoCSPTheta[i] = vitalbci_funcs.load_and_clfy(fullfile, ival, band / 4, classi)
        cvresultsAlpha[i], epoCSPAlpha[i] = vitalbci_funcs.load_and_clfy(fullfile, ival, band / 2, classi)
        cvresultsBeta[i], epoCSPBeta[i] = vitalbci_funcs.load_and_clfy(fullfile, ival, band, classi)
        cvresultsBroad[i], epoCSPBroad[i] = vitalbci_funcs.load_and_clfy(fullfile, ival, band * [0.5, 1], classi)
    maxres = np.max([np.mean(cvresultsAlpha[i]), np.mean(cvresultsBeta[i]), np.mean(cvresultsTheta[i]),
                     np.mean(cvresultsBroad[i])])
    print('Subject finished, maximum accuracy was %2.1f%% while original was %2.1f%%' % (100*maxres,
                                                                                     performance['acc_c'][0][i]))
AlphaWs = np.array([x.W for x in epoCSPAlpha])
BetaWs = np.array([x.W for x in epoCSPBeta])
ThetaWs = np.array([x.W for x in epoCSPTheta])
BroadWs = np.array([x.W for x in epoCSPBroad])
AlphaAs = np.array([x.A for x in epoCSPAlpha])
BetaAs = np.array([x.A for x in epoCSPBeta])
ThetaAs = np.array([x.A for x in epoCSPTheta])
BroadAs = np.array([x.A for x in epoCSPBroad])
Alphascores = np.array([x.score for x in epoCSPAlpha])
Betascores = np.array([x.score for x in epoCSPBeta])
Thetascores = np.array([x.score for x in epoCSPTheta])
Broadscores = np.array([x.score for x in epoCSPBroad])
Alphaspecs = np.array([x.spec for x in epoCSPAlpha])
Betaspecs = np.array([x.spec for x in epoCSPBeta])
Thetaspecs = np.array([x.spec for x in epoCSPTheta])
Broadspecs = np.array([x.spec for x in epoCSPBroad])

np.savez('/home/bbci/data/results/studies/vitalbci_season1/bbcpy_cv_results', cvresultsAlpha=cvresultsAlpha,
         cvresultsBeta=cvresultsBeta, cvresultsTheta=cvresultsTheta, cvresultsBroad=cvresultsBroad,
         AlphaWs=AlphaWs, AlphaAs=AlphaAs, Alphascores=Alphascores, Alphaspecs=Alphaspecs,
         BetaWs=BetaWs, BetaAs=BetaAs, Betascores=Betascores, Betaspecs=Betaspecs,
         ThetaWs=ThetaWs, ThetaAs=ThetaAs, Thetascores=Thetascores,Thetaspecs=Thetaspecs,
         BroadWs=BroadWs, BroadAs=BroadAs, Broadscores=Broadscores, Broadspecs=Broadspecs)



import numpy as np


class array(np.ndarray):
    def save(self, fname):
        dims = self.shape
        savelist = [ x.dataDict for x in np.nditer(self)]
        np.save(fname, np.reshape(savelist,dims))



import sklearn


# from functools import wraps


# class MakeEpochs(sklearn.base.BaseEstimator, sklearn.base.TransformerMixin):
#     def __call__(self, *args, **kwargs):
#         print('makey')
#         if isinstance(args[0], bbcpy.datatypes.eeg.Data):
#             self.fs = args[0].fs
#             self.ival = args[1]
#             outval, _ = self.transform(args[0], (args[0].mrk, args[0].mrk.y))
#             return outval
#         else:
#             raise TypeError('First argument needs to be of type Data')
#
#     def __init__(self, fs, ival):
#         self.fs = fs
#         self.ival = ival
#
#     def fit(self, X, y=None):
#         '''Nothing. for compatibility purposes'''
#         return self
#
#     # this is a deviation from sklearn standards, as the y-values are transformed as well.
#     # I (gabriel) wrote a custom pipeline for this, I'll try to get it into the official sklearn build
#     def transform(self, X, y):
#         """Make Epochs.
#
#         Parameters
#         ----------
#         X : ndarray, shape (n_matrices, n_channels, n_times)
#             Multi-channel time-series
#         y : ndarray, shape (n_trials, 2)
#             Marker positions and marker class.
#         Returns
#         -------
#         covmats : ndarray, shape (n_matrices, n_channels, n_channels)
#             Covariance matrices.
#         """
#         if isinstance(X, bbcpy.datatypes.eeg.Data):
#             X.y = y[1]
#             X.mrk = bbcpy.datatypes.eeg.Marker(y[0], y[1], X.mrk.className)
#             epo = bbcpy.functions.structural_manipulations.makeepochs(X, self.ival)
#             return epo, epo.y
#         else:
#             epo, epo_t = bbcpy.functions.structural_manipulations._makeepochs(X, self.fs, y[0], self.ival)
#             return epo, y[1]

class AutoTransformer(sklearn.base.BaseEstimator, sklearn.base.TransformerMixin):
    """

        :param func: function to import
        :param stdpars: standard parameters for the function call that can be set

        Parameters
        ----------
        func : function
            function to import
        stdpars : dict
            parameters for the function call becoming standard. Can be set as key-value pairs.

        Examples:
        _______
        var=bbcpy.pipeline.AutoTransformer(numpy.var, axis=0)

        hilbert=bbcpy.pipeline.AutoTransformer(scipy.signal.hilbert,  keepclass=True, axis=1)
        """

    # stdpars = {}
    # def __call__(self, *args, **kwargs):  # leave it although overwritten in __new__ so interpreter knows its callable
    #    return self.func(*args, **kwargs)

    # @classmethod
    def _get_param_names(self):  # needed for compatibility with sklearn
        return list(self.stdpars) + ['func']

    def get_stdpars(self):
        pars = self.get_params().copy()
        pars.pop('func')
        return pars

    # def __new__(cls, func, **stdpars):
    #     # Add stdpars to func.__doc___, here.
    #     tempdoc = "Transformer class of {0}:\n Runs {0} with {1:} by default\n".format(func.__name__, str(stdpars))+str(func.__doc__)
    #
    #     class AutoTrafo(cls):  # Create temporary class Func to be able to set __call__ doc correctly
    #         """ Transformer class of some function"""
    #
    #         __doc__ = tempdoc  # "Transformer class of {0}:\n{1}".format(func.__name__,tempdoc)
    #
    #         # wrap the doc of function to __call__
    #         #@wraps(func)
    #         def __call__(self, *args, **kwargs):
    #             kwargs2 = self.stdpars.copy()
    #             for key, value in kwargs.items():
    #                 kwargs2[key] = value  # reset to kwargs if passed
    #             if kwargs2.get('func'):  # need to remove func in order for call to work (work-around)
    #                 kwargs2.pop('func')
    #             #if keepclass:
    #             #    return args[0].__class__(func(*args, **kwargs2), *(
    #             #        args[0].__initargs__()))  # set __call__ to function,
    #                 # return outclass(func(*args, **kwargs2), *(
    #                 #    args[0].__initargs__()))  # set __call__ to function,
    #             #else:
    #             return func(*args, **kwargs2)  # set __call__ to function
    #         __call__.__doc__ = tempdoc
    #
    #         def __new__(cls2, func, keepclass=True, **stdpars2):
    #             return cls.__base__.__new__(cls2)
    #
    #         # def __init__(self, func, **stdpars2):
    #         #     self.stdpars = stdpars2
    #         #     for key, value in stdpars2.items():
    #         #         setattr(self, key, value)
    #
    #     obj = AutoTrafo(func, **stdpars)
    #     return obj

    def __init__(self, func, **stdpars):
        """Init."""
        self.func = func
        # stdpars['func'] = func
        self.stdpars = stdpars.keys()
        for key, value in stdpars.items():
            setattr(self, key, value)
        tempdoc = \
            "Transformer class of '{0}':\n Runs '{0}' with {1:} by default\n .transform evaluates the function\n " \
            ".fit doing nothing.\n \n Help of the function:".format(func.__name__, str(stdpars)) + str(func.__doc__)
        self.__doc__ = tempdoc

    def fit(self, x, y=None):
        """Nothing. for compatibility purposes"""
        return self

    def transform(self, x, y=None):
        """Make Epochs.

        Parameters
        ----------
        x : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series
        y : ndarray, shape (n_trials, 2)
            Marker positions and marker class.
        Returns
        -------
        covmats : ndarray, shape (n_matrices, n_channels, n_channels)
            Covariance matrices.
        """
        if y is not None:
            x.y = y
            x.mrk.y = y
        out = self.func(x, **self.get_stdpars())
        # if len(self.stdpars):
        #    out = self(x, **self.stdpars)
        #    #out = self(x, self.__getattribute__(*self.stdpars))
        # else:
        #    out = self(x)
        if y is None:
            return out
        elif hasattr(out, 'y'):
            return out, out.y
        else:
            return out, y



from bbcpy.pipeline.core import AutoTransformer
import warnings


# inherit from functools?!
class ImportFunc:
    """
        Import any function working on data to the bbcpy toolbox. These are automatically usable by the pipeline with
        an automated transformer type.
        :parameter
        func: function to import
        outClass: class that the function should return. Can be None to not care about the class, "same" for same class
         as input or a class type.
        **kwargs: arguments to the class
        Examples:
            var = ImportFunc(np.var,outClass=None
    """

    def __call__(self, data, *args, **kwargs):
        if not len(args) and not len(kwargs):
            args = self.args
            kwargs = self.kwargs
        if self.outClass is None:
            return self.__func__(data, *args, **kwargs)
        elif self.outClass == 'same':
            try:
                return data.__class__(self.__func__(data, *args, **kwargs), *data.__initargs__())
            except (TypeError):
                warnings.warn('Same class not possible as output, returning as Numpy array')
                return self.__func__(data, *args, **kwargs)
        else:
            if hasattr(self.outClass, '__initargs__'):
                return self.outClass(self.__func__(data, *args, **kwargs),
                                     *self.outClass.__initargs__())#to be improved: attribute check, if there
            else:
                return self.outClass(self.__func__(data, *args, **kwargs)) #to be improved: attribute check, if there

    def __init__(self, func,  *args, outClass=None, **kwargs):  # safe imports? give option for where to put in lib?
        self.__func__ = func
        self.args = args
        self.kwargs = kwargs
        self.outClass = outClass

    def trafo(self):
        return AutoTransformer(self.__func__, *self.args, **self.kwargs)



import matplotlib.pyplot as plt
import numpy as np
import bbcpy
import scipy.io as sio
from vitalbci_funcs import *

vp_fb_parameters = sio.loadmat('C:/data/results/studies/vitalbci_season1/vp_fb_parameters.mat')
classes = vp_fb_parameters['classes']

performance = sio.loadmat('C:/data/results/studies/vitalbci_season1/performance.mat')
data = np.load('C:/data/results/studies/vitalbci_season1/bbcpy_cv_results_broader.npz', allow_pickle=True)
for key in data.files:
    globals()[key] = data.get(key)
data.close()
Broaderspecs = Broadspecs  # correcting error in variable name

data = np.load('C:/data/results/studies/vitalbci_season1/bbcpy_cv_results.npz', allow_pickle=True)
for key in data.files:
    globals()[key] = data.get(key)
data.close()

dataset = bbcpy.load.eeg.dataset('C:/data/results/studies/vitalbci_season1/bbcpy_datsnip.npy')

# plot_and_compare(Broadscores, Broadspecs, performance)
# plot_and_compare(Alphascores, Alphaspecs, performance)
# plot_and_compare(Betascores, Betaspecs, performance)
# plot_and_compare(Thetascores, Thetaspecs, performance)

# plot_and_compare(Broaderscores, Broaderspecs, performance, f=np.arange(0, 101))
# plot_and_compare(Broader100scores, Broader100specs, performance, f=np.arange(0, 101))

# plot_and_compare_scalps(Alphascores, AlphaAs, dataset, classes, sorted=True)
# plot_and_compare_scalps(Betascores, BetaAs, dataset, classes, sorted=True)
# plot_and_compare_scalps(Thetascores, ThetaAs, dataset, classes, sorted=True)
# plot_and_compare_scalps(Broadscores, BroadAs, dataset, classes, sorted=True)

plot_and_compare_scalps_bands(Alphascores, Betascores, Thetascores, Broadscores, Broaderscores,
                              Broader100scores, AlphaAs, BetaAs, ThetaAs, BroadAs, BroaderAs,
                              Broader100As, Alphaspecs, Betaspecs, Thetaspecs, Broadspecs, Broaderspecs,
                              Broader100specs, dataset, classes, performance['acc_c'][0],
                              selection=np.random.default_rng().choice(len(dataset), size=10,
                                                                       replace=False))

neuronAP = np.load('../NEURONSomaAP.npz')
from scipy.signal import resample, resample_poly, decimate, welch

fs1 = 1000  # dataset[0].fs
fs2 = int(1 / (neuronAP['t'][1] - neuronAP['t'][0]))
y = resample_poly(neuronAP['V'] - neuronAP['V'][0], fs1, fs2)
y2 = resample(neuronAP['V'] - neuronAP['V'][0], int(fs1 / fs2 * neuronAP['V'].shape[0]))
y3 = decimate(neuronAP['V'] - neuronAP['V'][0], int(fs2 / fs1))
t = np.linspace(0, y.shape[0] / fs1, y.shape[0])
t2 = np.linspace(0, y2.shape[0] / fs1, y2.shape[0])
t3 = np.linspace(0, y3.shape[0] / fs1, y3.shape[0])
plt.figure()
plt.plot(neuronAP['t'], neuronAP['V'] - neuronAP['V'][0])
plt.plot(t, y)
plt.plot(t2, y2)
plt.plot(t3, y3)

yV = neuronAP['V'] - neuronAP['V'][0]
yV /= np.abs(yV).max()
tI, yI = IPSP(fs1, True)
tE, yE = EPSP(fs1, True)
tM, yM = MUAP(fs2, True)
plt.figure()
plt.plot(neuronAP['t'], yV)
plt.plot(tE, yE)
plt.plot(tI, yI)
plt.plot(tM, yM)
plt.xlim((0, 0.02))

f, pxx = welch(np.pad(yV, (0, int(fs2) - neuronAP['V'].shape[0])), fs2,
               nperseg=int(fs2), window='hamming', detrend=False)
plt.figure()
plt.semilogy(f, pxx * 10)
fy, pyy = welch(np.pad(y, (0, int(fs1) - y.shape[0])), fs1,
                nperseg=int(fs1), window='hamming', detrend=False)
# plt.plot(fy, pyy)
fyE, pyyE = welch(np.pad(yE, (0, int(fs1) - yE.shape[0])), fs1,
                  nperseg=int(fs1), window='hamming', detrend=False)
plt.semilogy(fyE, pyyE)
fyI, pyyI = welch(np.pad(yI, (0, int(fs1) - yI.shape[0])), fs1,
                  nperseg=int(fs1), window='hamming', detrend=False)
plt.semilogy(fyI, pyyI / 100)
fyM, pyyM = welch(np.pad(yM, (0, int(fs2) - yM.shape[0])), fs2,
                  nperseg=int(fs2), window='hamming', detrend=False)
plt.semilogy(fyM, pyyM / 5)
plt.xlim([0, 200])
plt.legend(['AP', 'EPSP', 'IPSP', 'EMG'])
plt.ylim([1e-9, 1e-6])

plt.figure()
plt.loglog(f, pxx)
# plt.loglog(fy, pyy)
plt.loglog(fyE, pyyE)
plt.loglog(fyI, pyyI)
plt.loglog(fyM, pyyM)
# plt.xlim([0, 2500])

specmodel = lambda f, E, I, A, M: (E * np.sqrt(pyyE[:f.shape[0]]) + I * np.sqrt(pyyI[:f.shape[0]]) +
                                   A * np.sqrt(pxx[:f.shape[0]]) + M * np.sqrt(pyyM[:f.shape[0]]))
from scipy import optimize

res = optimize.curve_fit(specmodel, fyE, np.sqrt(pyyE))  # santiy test
print(res[0])
testspec1 = 10 ** (Alphaspecs[0, 0, 0, :] / 20)
testspec2 = 10 ** (Alphaspecs[0, 1, 0, :] / 20)
f = np.arange(1, Alphaspecs.shape[3] + 1)
plt.figure()
plt.plot(f, testspec1)
plt.plot(f, testspec2)
plt.legend(['class0', 'class1'])

res = optimize.curve_fit(specmodel, f, testspec1 - testspec2, maxfev=100000)
plt.figure()
plt.plot(f, testspec1 - testspec2)
plt.plot(f, specmodel(f, *res[0]))

specmodelnonlin = lambda f, E, I, A, M, E2, I2, A2, M2: specmodel(f, E, I, A, M) + np.log(
    1 / specmodel(f, E2, I2, A2, M2) - 1)
resnonlin = optimize.curve_fit(specmodelnonlin, f, testspec1 - testspec2, maxfev=100000)
plt.figure()
plt.plot(f, testspec1 - testspec2)
plt.plot(f, specmodelnonlin(f, *resnonlin[0]))


def estimateSpec(f, spec1, spec2, pyyE, pyyI, pxx, startval=np.ones(5)):
    specnonlin = np.log(1 / spec1 - 1)
    logistic = lambda h: 1 / (1 + np.exp(-h))
    #EIspec = lambda EI: EI * np.sqrt(pyyE[:f.shape[0]]) + (1 - EI) * np.sqrt(pyyI[:f.shape[0]])
    EIspec = lambda EI: EI * np.sqrt(pyyE[:f.shape[0]]) + 1 / EI * np.sqrt(pyyI[:f.shape[0]])
    specmodel = lambda f, Amp, EI, A, EI2, Amp2: Amp * spec1 + EIspec(EI) + A * np.sqrt(pxx[:f.shape[0]]) + \
                                                 logistic(Amp2 * specnonlin + EI2)
    res = optimize.curve_fit(specmodel, f, spec2, startval, maxfev=100000)
    return res[0], specmodel(f, *res[0])


def estimateSpecwMuscle(f, spec1, spec2, pyyE, pyyI, pxx, pyyM, startval=np.ones(6)):
    specnonlin = np.log(1 / spec1 - 1)
    logistic = lambda h: 1 / (1 + np.exp(-h))
    EIspec = lambda EI: EI * np.sqrt(pyyE[:f.shape[0]]) + (1 - EI) * np.sqrt(pyyI[:f.shape[0]])
    specmodel = lambda f, Amp, EI, A, M, EI2, Amp2: Amp * spec1 + EIspec(EI) + A * np.sqrt(pxx[:f.shape[0]]) + \
                                                    M * np.sqrt(pyyM[:f.shape[0]]) + logistic(Amp2 * specnonlin + EI2)
    res = optimize.curve_fit(specmodel, f, spec2, startval, maxfev=100000)
    return res[0], specmodel(f, *res[0])


isubj = np.random.randint(Alphaspecs.shape[0])
f = np.arange(1, Alphaspecs.shape[3] + 1)
if Alphaspecs[isubj, 0, 0, 9] > Alphaspecs[isubj, 1, 0, 9]:
    testspec1 = 10 ** (Alphaspecs[isubj, 0, 0, :] / 20)
    testspec2 = 10 ** (Alphaspecs[isubj, 1, 0, :] / 20)
else:
    testspec2 = 10 ** (Alphaspecs[isubj, 0, 0, :] / 20)
    testspec1 = 10 ** (Alphaspecs[isubj, 1, 0, :] / 20)

res, estspec = estimateSpec(f, testspec1, testspec2, pyyE, pyyI, pxx, [1, 1, 0.1, 0.1, 0.1])

plt.figure()
plt.plot(f, 20 * np.log10(testspec1))
plt.plot(f, 20 * np.log10(testspec2))
plt.plot(f, 20 * np.log10(estspec))
plt.legend(['class0', 'class1', 'estimated'])
plt.title('%2.1f' % Alphascores[isubj, 0])



from bbcpy.functions.normalizations import baseline
import numpy as np
import pytest

def test_baseline():
    assert True


# WIP
# def test_baseline():
#     dummy_epo = np.ones((10, 16, 100))  # (samples x channels x epochs)
#     idxref = [20, 80]
#     time = np.array([range(np.int(idxref[0]),
#                            np.int(idxref[1])+1)])
#     epo_t = np.linspace(idxref[0], idxref[1], time.shape[1])
#
#     idxref_2 = [20, 120]
#
#     # raise if the idxref is out of the range
#     with pytest.raises(ValueError):
#         baseline(dummy_epo, idxref_2)
#
#     # Check if the mean is correct
#     epo_norm = baseline(dummy_epo, idxref)
#     t_epo_norm = np.ones((10, 16, 100))
#     t_epo_norm[:, :, epo_t] = 0
#     assert t_epo_norm == epo_norm



import scipy.io as sio

import bbcpy.load.eeg
import vitalbci_funcs
import numpy as np

performance = sio.loadmat('/home/bbci/data/results/studies/vitalbci_season1/performance.mat')
subdir_list = performance['subdir_list']
Nsubj = len(subdir_list)
datsnip = [dict] * Nsubj
for i, fname in enumerate(subdir_list):
    fname = fname[0][0]
    subjname = fname.split('_')[0]
    fullfile = '/home/bbci/data/bbciMat/' + fname + '/imag_arrow' + subjname + '_250Hz.mat'  # _250Hz option
    data = bbcpy.load.eeg.bbci_mat(fullfile, load_data=False)
    datsnip[i] = data.dataDict()

np.save('/home/bbci/data/results/studies/vitalbci_season1/bbcpy_datsnip', datsnip)



import warnings

import numpy as np
import scipy as sp
from matplotlib import pyplot as plt


def map(eegdata, v=None, clim='minmax', cb_label='', colorbar=True, senspos=True, aspect='equal',
        extent=None):
    '''
    Usage:
        scalpmap(mnt, v, clim='minmax', cb_label='')
    Parameters:
        mnt: a 2D array of channel coordinates (channels x 2)
        v:   a 1D vector (channels)
        clim: limits of color code, either
          'minmax' to use the minimum and maximum of the data
          'sym' to make limits symmetrical around zero, or
          a two element vector giving specific values
        cb_label: label for the colorbar
    '''
    mnt = eegdata.chans.mnt
    if np.any(np.isnan(mnt)):
        v = v[~np.isnan(mnt[:, 0])]
        mnt = mnt[~np.isnan(mnt[:, 0]), :]
        warnings.warn('Some sensor positions are undefined and thus excluded from plotting.')

    maxx = np.max([mnt[:, 0].max(), 1])
    maxy = np.max([mnt[:, 1].max(), 1])
    maxi = np.max([maxx,maxy])
    if extent is None:
        extent = np.array([-1, 1, -1, 1]) * maxi
    # interpolate between channels
    xi, yi = np.linspace(-1, 1, 100) * maxx, np.linspace(-1, 1, 100) * maxy
    xi, yi = np.meshgrid(xi, yi)
    rbf = sp.interpolate.Rbf(mnt[:, 0], mnt[:, 1], v, function='linear')
    zi = rbf(xi, yi)

    # mask area outside of the scalp
    a, b, n, r = 50, 50, 100, 50
    mask_y, mask_x = np.ogrid[-a:n - a, -b:n - b]
    mask = mask_x * mask_x + mask_y * mask_y >= r * r
    zi[mask] = np.nan

    if clim == 'minmax':
        vmin = v.min()
        vmax = v.max()
    elif clim == 'sym':
        vmin = -np.absolute(v).max()
        vmax = np.absolute(v).max()
    else:
        vmin = clim[0]
        vmax = clim[1]

    plt.imshow(zi, vmin=vmin, vmax=vmax, aspect=aspect, origin='lower', extent=extent, cmap='jet')
    if colorbar:
        plt.colorbar(shrink=.5, label=cb_label)
    if senspos:
        plt.scatter(mnt[:, 0], mnt[:, 1], c='k', marker='+', vmin=vmin, vmax=vmax)
    plt.axis('off')



# import pyriemann as pr
import numpy as np
import sklearn
import scipy as sp
import scipy.signal
import warnings
import fnmatch

import bbcpy


class Marker(np.ndarray):
    def __new__(cls, mrk_pos, mrk_class, mrk_class_name=None, mrk_fs=1):
        obj = np.asarray(mrk_pos).view(cls)
        obj.y = np.array(mrk_class)
        if mrk_class_name is None:
            mrk_class_name = np.unique(mrk_class)
        obj.className = mrk_class_name
        obj.fs = mrk_fs
        return obj

    def __array_wrap__(self, out_arr, context=None):  #
        obj = super().__array_wrap__(out_arr, context)
        obj.y = np.copy(self.y)
        obj.className = np.copy(self.className)
        obj.fs = np.copy(obj.fs)
        return obj

    def __array_finalize__(self, obj):  #
        self.y = getattr(obj, 'y', None)
        self.className = getattr(obj, 'className', None)
        self.fs = getattr(obj, 'fs', None)

    def __init__(self, mrk_pos, mrk_class, mrk_class_name=None, mrk_fs=1):
        return  # super().init()

    def __getitem__(self, key):
        newy = self.y[key].copy()
        if isinstance(key, int) or ((not isinstance(key, slice)) and len(key) == 1):
            newy = [newy]
        leftclasses = np.unique(self.y[key])
        if len(leftclasses) < len(self.className):
            # if not isinstance(key, int) and len(key) > 1:
            #    warnings.warn('removing void classes')
            newind = 0
            for i in range(len(self.className)):
                if i in leftclasses:
                    newy[newy == i] = newind
                    newind += 1
        newclassName = np.copy([self.className[lc] for lc in leftclasses])
        obj = Marker(super().__getitem__(key).copy(), newy, newclassName, mrk_fs=np.copy(self.fs))
        # add channel selection!
        # if len(self.shape):
        #    obj = Marker(super().__getitem__(key).copy(), newy, newclassName, mrk_fs=np.copy(self.fs))
        # else:
        #    obj = Marker(self.copy(), newy, newclassName, mrk_fs=np.copy(self.fs))
        return obj

    def sort(self, axis=-1, kind=None, order=None):
        sortedinds = self.argsort(axis, kind, order)
        return self[sortedinds]

    def in_ms(self):
        return self / self.fs * 1000

    def in_samples(self, fs=None):
        if fs is None:
            return self
        else:
            return self / self.fs * fs

    def select_classes(self, classes):
        if (type(classes) == list) & (type(classes[0]) == str):
            classes_int = []
            for ii in classes:
                inds = np.where(self.className == ii)[0]
                if inds >= 0:
                    classes_int.append(inds)
                else:
                    warnings.warn('Class not found: "%s"' % ii)
            classes = classes_int
        if type(classes) == int:
            selected = self.y == classes
        else:
            selected = np.zeros(self.shape, dtype=bool)
            for ii in classes:
                selected |= self.y == ii
        obj = self[selected]
        return obj


class Chans(list):
    mnt = None
    shape = ()

    def __init__(self, clab, mnt):
        super().__init__(clab)
        self.mnt = mnt
        self.shape = (len(self),)

    def __initargs__(self):
        return self.mnt

    def __getitem__(self, key):
        if isinstance(key, (list, tuple, np.ndarray)) and len(key) > 1:
            if any([isinstance(k, str) for k in key]):
                key = self.index(key)
            templist = []  # needed because super() does not work inside list comprehension
            for k in key:
                templist.append(super().__getitem__(k))
            obj = self.__class__(templist, self.mnt[key])
        else:
            if isinstance(key, (list, tuple, np.ndarray)) and len(key) == 1:
                key = key[0]
            if isinstance(key, str):
                obj = self.__getitem__(self.index(key))
            else:
                obj = self.__class__(super().__getitem__(key),
                                     self.mnt[key])  # problematic selecting one makes list of char
        obj.shape = (len(obj),)
        return obj

    def copy(self):
        return self.__class__(super().copy(), self.mnt.copy())

    def index(self, __value, __start=..., __stop=..., markmissing=False):
        if isinstance(__value, (list, np.ndarray, tuple)):
            isstr = [isinstance(k, str) for k in __value]
            if any(isstr) and sum(isstr) < len(__value):
                raise ValueError('Either all indices have to be Strings or none.')
            if any(isstr):
                isnot = [x[0] == '~' for x in __value]
                tempChan = self.copy()
                if any(isnot):
                    for i in reversed(np.where(isnot)[0]):
                        tempChan = tempChan[tempChan.index(__value[i])]
                        __value.pop(i)
                if len(__value):  # still commands left after exclusions
                    tempinds = [tempChan.index(name) for name in __value]
                    for i, inds in enumerate(tempinds):
                        if isinstance(inds, int):
                            tempinds[i] = [inds]
                    tempinds = np.unique([item for sublist in tempinds for item in sublist])
                    return [self.index(k) for k in tempChan[tempinds]]
                else:  # only exlcusions
                    tempChan = np.unique(tempChan)
                    return [self.index(k) for k in tempChan]
            else:
                return __value
        else:
            if isinstance(__value, (int, slice)):
                return __value
            keys = __value.split(',')
            if len(keys) > 1:
                return self.index(keys)
            __value = __value.strip()
            if __value[0] == '~':
                exclude = True
                __value = __value[1:]
            else:
                exclude = False
            items = fnmatch.filter(self, __value)
            if (len(items) == 0):
                if markmissing:
                    return None
                else:
                    raise ValueError('\'%s\' not found in channels' % __value)
            if exclude:
                allits = self.copy()
                [allits.remove(it) for it in items]
                items = allits
            outlist = []
            for it in items:
                outlist.append(super().index(it))
            if len(outlist) == 1:
                return outlist[0]
            else:
                return outlist


class Data(np.ndarray):
    def nT(self):
        if np.isscalar(self):
            return 1
        else:
            return self.shape[1]

    def nCh(self):
        if np.isscalar(self):
            return 1
        else:
            return self.shape[0]

    def __new__(cls, data, fs, mrk=None, chans=None):
        obj = np.asarray(data).view(cls)
        obj.fs = fs
        obj.mrk = mrk
        obj.chans = chans
        return obj

    # def __init__(self, data, fs, mrk=None, chans=None):
    # data = np.array(data)
    #    return

    def __array_wrap__(self, out_arr, context=None, *args, **kwargs):  # used for printing,squeezing etc
        # return self.__class__(self, *self.__initargs__())
        obj = super().__array_wrap__(out_arr, context)
        obj.fs = self.fs.copy()
        obj.mrk = self.mrk.copy()
        obj.chans = self.chans.copy()
        return obj

    def __array_finalize__(self, obj):  #
        if obj is None: return
        self.fs = getattr(obj, 'fs', None)
        self.mrk = getattr(obj, 'mrk', None)
        self.chans = getattr(obj, 'chans', None)

    def __initargs__(self):
        return self.fs, self.mrk, self.chans

    def __getitem__(self, key):
        if isinstance(key, type(np.newaxis)):
            chans = self.chans.copy()
        elif isinstance(key, (int, str, slice, list, np.ndarray)) or (len(key) == 1):  # or \
            #        (isinstance(key, np.ndarray) and key.dtype == bool):
            key = self.chans.index(key)
            chans = self.chans.copy()[key]
        elif isinstance(key, tuple) and (isinstance(key[0], (int, str, slice, list, np.ndarray))) and (
                len(self.chans.shape) > 0):  # or \
            # (isinstance(key[0], np.ndarray) and key[0].dtype == bool):
            key = list(key)
            key[0] = self.chans.index(key[0])
            key = tuple(key)
            chans = self.chans.copy()[key[0]]
        else:
            chans = self.chans.copy()
        if isinstance(key, tuple) and len(key) > 1 and (isinstance(key[1], (str, list, np.ndarray))):
            key = list(key)
            key[1] = gettimeindices(key[1], self.fs)
            key = tuple(key)

        x = super().__getitem__(key).copy()
        if isinstance(key[1], slice):
            fs = self.fs / key[1].step
        else:
            fs = self.fs
        if x.ndim < 2:  # dimension is lost, keep dimensionnality ?
            if isinstance(key, tuple) and (isinstance(key[0], int) or len(key[0]) == 1):
                x = x[np.newaxis]
            else:
                x = x[:, np.newaxis]
        obj = Data(x, fs, self.mrk, chans)
        return obj

    def copy(self):
        obj = self.__class__(self, *self.__initargs__())
        obj.fs = self.fs.copy()
        obj.mrk = self.mrk.copy()
        obj.chans = self.chans.copy()
        return obj

    def epochs(self, ival, mrk=None, clsinds=None):
        if mrk is None:
            try:
                mrk = self.mrk
            except (NameError, AttributeError):
                raise AttributeError('Error: Cannot make epochs: mrk not provided.')
        if clsinds is not None:
            mrk = mrk.select_classes(clsinds)
        ival = gettimeindices(ival, 1000)
        epo = bbcpy.functions.structural.makeepochs(self, ival, mrk)
        return epo

    def lfilter(self, band, order=5, filttype='*', filtfunc=sp.signal.butter):
        return bbcpy.functions.temporal.lfilter(self, band, order=order, filttype=filttype, filtfunc=filtfunc)
        # band = np.array(band)
        # if filttype == '*':
        #     if band.shape == (2,):
        #         filttype = 'bandpass'
        #     else:
        #         filttype = 'low'
        # [b, a] = filtfunc(order, band / self.fs * 2, filttype)
        # return Data(sp.signal.lfilter(b, a, self, axis=-1), *self.__initargs__())

    # @wraps(bbcpy.functions.statistics.cov)  #does not work due to circular import no solution found so far
    def cov(self, **kwargs):
        return bbcpy.functions.statistics.cov(self, **kwargs)

    def pca(self, **kwargs):
        pca_fitter = sklearn.decomposition.PCA(**kwargs)
        pca_fitter = pca_fitter.fit(self)
        # data = pca_fitter.transform(self)
        data = Data(pca_fitter.transform(self), self.fs, self.mrk)
        data.pcaobj = pca_fitter  # all values like W, A, d are stored in pca object of the EEGdata object
        # self.flat = data.flatten()
        return data

    def save(self, fname, saveData=True):
        np.savez(fname, **self.dataDict(withData=saveData))
        # warnings.warn('Unfinished saving procedure.') could not find what was missing.

    def append(self, data, axis=-1):
        obj = np.concatenate([self, data], axis=axis)
        if axis in (1, -1):  # append in time
            if self.fs != data.fs:
                raise ValueError('Sampling rates must match!')
            if len(self.chans) != len(data.chans) or (
                    np.any(np.isnan([self.chans.index(chan, markmissing=True) for chan in data.chans]))):
                raise ValueError('Channels must match, they do not.')
            mrk = bbcpy.datatypes.eeg.Marker(np.concatenate([self.mrk, data.mrk + self.shape[-1] / self.fs]),
                                             np.concatenate([self.mrk.y, data.mrk.y]), self.mrk.className, self.mrk.fs)
            return bbcpy.datatypes.eeg.Data(obj, self.fs, mrk, self.chans)
        else:  # append in channels
            raise NotImplementedError('not implemented, yet, sorry')

    def dataDict(self, withData=True):
        if withData:
            if self.mrk is None:
                return {'X': np.asarray(self), 'fs': self.fs, 'clab': np.asarray(self.chans), 'mnt': self.chans.mnt}
            return {'X': np.asarray(self), 'fs': self.fs, 'clab': np.asarray(self.chans), 'mnt': self.chans.mnt,
                    'mrk_pos': np.asarray(self.mrk), 'mrk_class': self.mrk.y, 'mrk_className': self.mrk.className}
        else:
            if self.mrk is None:
                return {'fs': self.fs, 'clab': np.asarray(self.chans), 'mnt': self.chans.mnt}
            return {'fs': self.fs, 'clab': np.asarray(self.chans), 'mnt': self.chans.mnt,
                    'mrk_pos': np.asarray(self.mrk), 'mrk_class': self.mrk.y, 'mrk_className': self.mrk.className}


class Epo(Data):
    def __new__(cls, data, time, fs, mrk, chans=None):
        obj = np.asarray(data).view(cls)
        obj.fs = fs
        obj.y = mrk.y
        obj.className = mrk.className
        obj.mrk = mrk
        obj.chans = chans
        obj.t = time
        return obj

    # def __init__(self, data, time, fs, mrk, chans=None):
    # data = np.array(data)
    # assert len(data.shape) < 4

    def __initargs__(self):
        return self.t, self.fs, self.mrk, self.chans

    def __array_wrap__(self, out_arr, context=None, *args, **kwargs):
        obj = super().__array_wrap__(out_arr, context, *args, **kwargs)
        obj.y = np.copy(self.y)
        obj.className = np.copy(self.className)
        return obj

    def __array_finalize__(self, obj):  #
        if obj is None: return
        self.fs = getattr(obj, 'fs', None)
        self.t = getattr(obj, 't', None)
        self.y = getattr(obj, 'y', None)
        self.className = getattr(obj, 'className', None)
        self.mrk = getattr(obj, 'mrk', None)
        self.chans = getattr(obj, 'chans', None)

    def __getitem__(self, key):
        if isinstance(key, type(np.newaxis)):
            mrk = self.mrk.copy()
            choosingmrk = True
        elif isinstance(key, (int, slice, list, np.ndarray)) or (len(key) == 1):  # or \
            #        (isinstance(key, np.ndarray) and key.dtype == bool):
            mrk = self.mrk.__getitem__(key).copy()
            choosingmrk = isinstance(key, slice) and slice.start == slice.stop or isinstance(key, int) or \
                          isinstance(key, (list, tuple)) and (len(key) == 1 or key.dtype == bool and key.sum() == 1)
        elif isinstance(key, tuple) and (isinstance(key[0], (int, slice, list, tuple, np.ndarray))):  # or \
            # (isinstance(key[0], np.ndarray) and key[0].dtype == bool):
            mrk = self.mrk.__getitem__(key[0]).copy()
            choosingmrk = isinstance(key[0], slice) and slice.start == slice.stop or isinstance(key[0], int) or \
                          isinstance(key[0], (list, tuple, np.ndarray)) and \
                          (len(key[0]) == 1 or key[0].dtype == bool and key[0].sum() == 1)
        else:
            mrk = self.mrk.copy()
            choosingmrk = False
        if isinstance(key, tuple) and len(key) > 1 and (isinstance(key[1], (int, str, slice, list, tuple, np.ndarray))) \
                and (len(self.chans.shape) > 0):
            key = list(key)
            if isinstance(key[1], (list, tuple, np.ndarray)) and len(key[1]) == 1:
                key[1] = key[1][0]
            key[1] = self.chans.index(key[1])
            key = tuple(key)
            chans = self.chans.copy()[key[1]]
        else:
            chans = self.chans.copy()
        if isinstance(key, tuple) and len(key) > 2 and (isinstance(key[2], str) or
                                                        (isinstance(key[2], (list, np.ndarray)) and any(
                                                            [isinstance(k, str) for k in key[2]]))):
            key = list(key)
            key[2] = gettimeindices(key[2], 1000)
            # recalculate according to t
            if isinstance(key[2], slice):
                if key[2].start is None:
                    start = None
                else:
                    start = np.where(self.t >= key[2].start)[0][0]
                if key[2].stop is None:
                    stop = None
                else:
                    stop = np.where(self.t >= key[2].stop)[0][0]
                if key[2].step is None:
                    step = None
                else:
                    if np.round(key[2].step / 1000 * self.fs) < 1:
                        raise ValueError('Step size too small.')
                    step = int(np.round(key[2].step / 1000 * self.fs))
                key[2] = slice(start, stop, step)
            else:
                for i, k in enumerate(key[2]):
                    if k < self.t[0] or k > self.t[-1]:
                        raise ValueError('indexed timepoint %f is out of range of data' % k)
                    key[2][i] = np.argmin(np.abs(k - self.t))
                tempkey = np.unique(key[2])
                if len(tempkey) < len(key[2]):
                    warnings.warn('Some indices where used multiple times, reducing to unique indices. '
                                  'This might be caused by sampling rate issues')
                key[2] = tempkey
            key = tuple(key)
        x = np.ndarray.__getitem__(self, key).copy()  # super() would be Data thats why ndarray
        if isinstance(key, tuple) and len(key) > 2:
            t = self.t[key[2]]
            if isinstance(key[2], slice) and key[2].step is not None:
                fs = float(self.fs) / key[2].step
            else:
                fs = self.fs
        else:
            t = self.t
            fs = self.fs

        if x.ndim < 3 and choosingmrk:
            x = x[np.newaxis]
        obj = Epo(x, t, fs, mrk, chans)
        return obj

    # def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
    # print('Hello ufunc')
    #    return super().__array_ufunc__(ufunc, method, np.asarray(*inputs), **kwargs)
    # return self.__class__(super().__array_ufunc__(ufunc, method, np.asarray(*inputs), **kwargs), *self.__initargs__())
    #    if method == '__call__':
    #        return self.__class__(ufunc(*inputs, **kwargs), self.__initargs__())
    #    else:
    #        return ufunc(*inputs, **kwargs)

    # def __array_function__(self, ufunc, method, *inputs, **kwargs):
    #    print('Hello array function')
    #    if method=='__call__':
    #        return self.__class__(ufunc(*inputs, **kwargs), self.__initargs__())
    #    else:
    #        return ufunc(*inputs, **kwargs)

    def dataDict(self, withData=True):
        if withData:
            if self.mrk is None:
                return {'X': np.asarray(self), 'fs': self.fs, 'time': self.t, 'clab': np.asarray(self.chans),
                        'mnt': self.chans.mnt}
            return {'X': np.asarray(self), 'fs': self.fs, 'time': self.t, 'clab': np.asarray(self.chans),
                    'mnt': self.chans.mnt,
                    'mrk_pos': np.asarray(self.mrk), 'mrk_class': self.mrk.y, 'mrk_className': self.mrk.className}
        else:
            if self.mrk is None:
                return {'fs': self.fs, 'clab': np.asarray(self.chans), 'mnt': self.chans.mnt}
            return {'fs': self.fs, 'clab': np.asarray(self.chans), 'mnt': self.chans.mnt,
                    'mrk_pos': np.asarray(self.mrk), 'mrk_class': self.mrk.y, 'mrk_className': self.mrk.className}

    def nT(self):
        if np.isscalar(self):
            return 1
        else:
            return self.shape[2]

    def nCh(self):
        if np.isscalar(self):
            return 1
        else:
            return self.shape[1]

    def nEpo(self):
        if np.isscalar(self):
            return 1
        else:
            return self.shape[0]

    def nClass(self):
        if np.isscalar(self):
            return 1
        else:
            return len(self.mrk.className)

    def lfilter(self, band, order=5, filttype='*', filtfunc=sp.signal.butter):
        band = np.array(band)
        if len(band.shape):
            assert band.shape == (2,)
        warnings.warn(
            'Filtering the epoched data is not optimal due to filter artefacts. Consider filtering the continous data before segmentation.')
        return super().lfilter(band, order=order, filttype=filttype, filtfunc=filtfunc)

    def classmean(self, classid=None):
        """"""
        if classid == None:
            outdata = np.empty((self.nClass(), *self.shape[1:]))
            for i in range(self.nClass()):
                outdata[i] = self[self.y == i].mean(axis=0)
            return Epo(outdata, *self.__initargs__())
        return self[self.y == classid]


def gettimeindices(orig_key, fs):
    if isinstance(orig_key, (list, np.ndarray)):
        key = orig_key
        for i in range(len(orig_key)):
            key[i] = gettimeindices(orig_key[i], fs)
            if isinstance(key[i], slice):
                if len(key) > 1:
                    raise ValueError('It is not possible to combine multiple indexings if any of them is a slice.')
                return key[i]
        newkey = []
        for sublist in key:
            if isinstance(sublist, (list, np.ndarray)):
                for item in sublist:
                    newkey.append(item)
            else:
                newkey.append(sublist)
        key = newkey

    elif isinstance(orig_key, str):  # str "100ms:450ms" or "100ms,230ms,..."
        slice_found = False
        key = orig_key.split(",")
        for i, k in enumerate(key):
            key[i] = k.split(":")
            if isinstance(key[i], list) and len(key) > 1 and len(key[i]) > 1:
                raise ValueError('It is not possible to combine multiple indexings if any of them is a slice.')

        key = [[k.strip() for k in k2] for k2 in key]
        for i, k in enumerate(key):
            for i2, k2 in enumerate(k):
                # inums = np.sum([s.isnumeric() for s in k2]) does not work for floats due to point
                inums = len(k2) - np.sum([s.isalpha() for s in k2])
                num = float(k2[:inums])
                if inums < len(k2):
                    unit = k2[inums:]
                    if unit == 'ms':
                        factor = float(fs) / 1000
                    elif unit in ('s', 'sec'):
                        factor = float(fs)
                    elif unit in ('m', 'min'):
                        factor = float(fs) * 60
                    elif unit == 'h':
                        factor = float(fs) * 3600
                    k[i2] = int(np.round(num * factor))
                else:
                    k[i2] = int(k2)
                    warnings.warn(
                        'No unit given for one or more elements in [%s], assuming samples for these.' % (orig_key))
            if len(k) > 1:
                key = slice(*key[i])
                slice_found = True
        if not slice_found:
            key = [item for sublist in key for item in sublist]
    else:
        key = orig_key
    return key



import numpy as np
import scipy as sp
import scipy.io
import scipy.signal


def music(patt, L, grid):
    """Performs a MUSIC-scan for a given pattern in the leadfield

    Parameters
    ----------
    patt : (N,M) ndarray for N channels;
        each column in patt represents a spatial pattern;
        (only the span(patt) matters; mixing of the patterns has no effect)
    L : (N,M,3) ndarray for m grid points; L(:,i,j) is the potential
        of a unit dipole at point i in direction j
    grid : (M,3) ndarray denoting locations of grid points

    Returns
    -------
    s : (M, K)
        s(i,k) indicates fit-quality (from 0 (worst) to 1 (best)) at grid-point
        i of k.th dipole (i.e. preceeding k-1 dipoles are projected out
        at each location); the first column is the 'usual' MUSIC-scan
    vmax : (N, ) ndarray
        the field of the best dipole
    imax : int
        denotes grid-index of best dipole
    dip_mom : (3, ) ndarray
        the moment of the  best dipole
    dip_loc : (3, ) ndarray
        the location of the  best dipole

    """
    if np.ndim(patt) == 1:
        patt = np.expand_dims(patt, 1)
    nchan, nx = patt.shape

    dims = np.ones(4, dtype=int)
    dims[:len(L.shape)] = L.shape
    nchan, ng, ndum, nsubj = dims

    data = patt / np.linalg.norm(patt)
    nd = np.minimum(nx, ndum)

    [s, vmax, imax] = calc_spacecorr_all(L, data, nd)

    dip_mom = vmax2dipmom(L, imax, vmax)
    dip_loc = grid[imax, :]

    return s, vmax, imax, dip_mom, dip_loc


def calc_spacecorr_all(V, data, nd):
    dims = np.ones(4, dtype=int)
    dims[:len(V.shape)] = V.shape
    nchan, ng, ndum, nsubj = dims  # [nchan,ng,ndum]=size(V);
    s = np.zeros((ng, nd))  # s=zeros(ng,nd);

    for i in range(ng):  # for i=1:ng;
        Vortholoc = sp.linalg.orth(np.squeeze(V[:, i, :]))  # Vortholoc=orth(squeeze(V(:,i,:)));
        s[i, :] = calc_spacecorr(Vortholoc, data, nd)  # s(i,:)=calc_spacecorr(Vortholoc,data,nd);

    imax = np.argmax(s[:, 0])  # [smax,imax]=max(s(:,1));
    Vortholoc = sp.linalg.orth(np.squeeze(V[:, imax, :]))  # Vortholoc=orth(squeeze(V(:,imax,:)));
    vmax, sbest = calc_bestdir(Vortholoc, data)  # vmax=calc_bestdir(Vortholoc,data);
    return s, vmax, imax


def calc_spacecorr(Vloc, data_pats, nd):
    A = np.matmul(data_pats.T, Vloc)  # A=data_pats'*Vloc;
    s = np.sqrt(np.abs(np.linalg.eigvals(np.outer(A, A))[:nd]))  # s=sd(1:nd);

    return s


def vmax2dipmom(V, imax_all, vmax_all):
    if np.isscalar(imax_all):
        ns = 1
        vmax_all = np.expand_dims(vmax_all, 1)
        imax_all = np.expand_dims(imax_all, 0)
    else:
        ns = len(imax_all)

    dips_mom_all = np.zeros((ns, 3))  # dips_mom_all = zeros(ns, 3);

    for i in range(ns):  # for i=1:ns
        Vloc = np.squeeze(V[:, imax_all[i], :])  # Vloc = squeeze(V(:, imax_all(i),:));

        v = vmax_all[:, i]  # v = vmax_all(:, i);
        dip = np.linalg.inv(Vloc.T @ Vloc) @ Vloc.T @ v  # dip = inv(Vloc'*Vloc)*Vloc' * v;
        dips_mom_all[i, :] = dip.T / np.linalg.norm(dip)  # dips_mom_all(i,:)=dip'/norm(dip);

    return dips_mom_all


def calc_bestdir(Vloc, data_pats, proj_pats={}):
    if len(proj_pats) == 0:  # if nargin==2
        A = data_pats.T @ Vloc  # A=data_pats'*Vloc;
        u, s, v = sp.linalg.svd(A)  # [u s, v]=svd(A);
        vmax = Vloc @ v[:, 0]  # vmax=Vloc*v(:,1);
        vmax = vmax / np.linalg.norm(vmax)  # vmax=vmax/norm(vmax);
        s = s[0]  # s=s(1,1);
    else:
        n, m = Vloc.shape  # [n m]=size(Vloc);
        a = proj_pats.T @ Vloc
        V_proj = sp.linalg.orth(Vloc - proj_pats @ a)  # V_proj=orth(Vloc-proj_pats*(proj_pats'*Vloc));
        A = data_pats.T @ V_proj  # A=data_pats'*V_proj;
        u, s, v = sp.linalg.svd(A)  # [u, s v]=svd(A);
        BB = Vloc.T @ proj_pats  # BB=(Vloc'*proj_pats);
        Q = np.linalg.inv(np.identity(m) - BB @ BB.T + np.sqrt(np.finfo(float).eps))  # Q=inv(eye(m)-BB*BB'+sqrt(eps));
        vmax = Vloc @ (Q @ Vloc.T @ (V_proj @ v[:, 0]))  # vmax=Vloc*(Q*Vloc'*(V_proj*v(:,1)));
        vmax = vmax / np.linalg.norm(vmax)  # vmax=vmax/norm(vmax);
        s = s[0, 0]  # s=s(1,1);

    return vmax, s




from setuptools import setup
from setuptools import find_packages

# read the contents of your README file for long_description
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(name='bbcpy',
      version='0.1.1',
      description='A novel Python BCI toolbox',
      url='https://github.com/bbcpy/bbcpy',
      author='Neurotechnology Group TU Berlin',
      author_email='bbcpy@tu-berlin.de',
      license='MIT',
      packages=find_packages(),
      install_requires=[
          'scikit-learn', 'pyriemann', 'numpy', 'scipy', 'matplotlib', ''
      ],
      test_suite='nose.collector',
      tests_require=['nose'],
      extras_require=dict(tests=['pytest']),
      zip_safe=False,
      long_description=long_description,
      long_description_content_type='text/markdown'
      )
